from __future__ import annotations

from typing import cast

from coredis._utils import nativestr
from coredis.response._callbacks import ResponseCallback
from coredis.response.types import ClusterNode, ClusterNodeDetail
from coredis.typing import (
    AnyStr,
    Mapping,
    RedisValueT,
    ResponsePrimitive,
    StringT,
)


class ClusterLinksCallback(
    ResponseCallback[list[dict[AnyStr, ResponsePrimitive]], list[dict[AnyStr, ResponsePrimitive]]]
):
    def transform(
        self,
        response: list[dict[AnyStr, ResponsePrimitive]],
    ) -> list[dict[AnyStr, ResponsePrimitive]]:
        return response


class ClusterInfoCallback(ResponseCallback[StringT, dict[str, str]]):
    def transform(
        self,
        response: StringT,
    ) -> dict[str, str]:
        response_str = nativestr(response)
        return dict([line.split(":") for line in response_str.splitlines() if line])


class ClusterSlotsCallback(
    ResponseCallback[
        list[list[ResponsePrimitive | list[ResponsePrimitive | dict[StringT, ResponsePrimitive]]]],
        dict[tuple[int, int], tuple[ClusterNode, ...]],
    ]
):
    def transform(
        self,
        response: list[
            list[ResponsePrimitive | list[ResponsePrimitive | dict[StringT, ResponsePrimitive]]]
        ],
    ) -> dict[tuple[int, int], tuple[ClusterNode, ...]]:
        res: dict[tuple[int, int], tuple[ClusterNode, ...]] = {}

        for slot_info in response:
            min_slot = cast(int, slot_info[0])
            max_slot = cast(int, slot_info[1])
            nodes = [
                cast(tuple[StringT, int, StringT], tuple(node[:3]))
                for node in slot_info[2:]
                if isinstance(node, list)
            ]
            res[(min_slot, max_slot)] = tuple(self.parse_node(node) for node in nodes)
            res[(min_slot, max_slot)][0]["server_type"] = "master"

        return res

    def parse_node(self, node: tuple[StringT, int, StringT | None]) -> ClusterNode:
        return ClusterNode(
            host=nativestr(node[0]),
            port=int(node[1]),
            node_id=nativestr(node[2]) if len(node) > 2 else "",
            server_type="slave",
        )


class ClusterNodesCallback(ResponseCallback[StringT | list[StringT], list[ClusterNodeDetail]]):
    def transform(
        self,
        response: StringT | list[StringT],
    ) -> list[ClusterNodeDetail]:
        resp: list[str] | str

        if isinstance(response, list):
            resp = [nativestr(row) for row in response]
        else:
            resp = nativestr(response)
        current_host = nativestr(self.options.get("current_host", ""))

        def parse_slots(s: str) -> tuple[list[int], list[dict[str, RedisValueT]]]:
            slots: list[int] = []
            migrations: list[dict[str, RedisValueT]] = []

            for r in s.split(" "):
                if "->-" in r:
                    slot_id, dst_node_id = r[1:-1].split("->-", 1)
                    migrations.append(
                        {
                            "slot": int(slot_id),
                            "node_id": dst_node_id,
                            "state": "migrating",
                        }
                    )
                elif "-<-" in r:
                    slot_id, src_node_id = r[1:-1].split("-<-", 1)
                    migrations.append(
                        {
                            "slot": int(slot_id),
                            "node_id": src_node_id,
                            "state": "importing",
                        }
                    )
                elif "-" in r:
                    start, end = r.split("-")
                    slots.extend(list(range(int(start), int(end) + 1)))
                else:
                    slots.append(int(r))

            return slots, migrations

        if isinstance(resp, str):
            resp = resp.splitlines()

        nodes: list[ClusterNodeDetail] = []

        for line in resp:
            parts = line.split(" ", 8)
            (
                self_id,
                addr,
                flags,
                master_id,
                ping_sent,
                pong_recv,
                _,
                link_state,
            ) = parts[:8]

            host, port = addr.rsplit(":", 1)

            node: ClusterNodeDetail = {
                "id": self_id,
                "host": host or current_host,
                "port": int(port.split("@")[0]),
                "flags": tuple(flags.split(",")),
                "master": master_id if master_id != "-" else None,
                "ping_sent": int(ping_sent),
                "pong_recv": int(pong_recv),
                "link_state": link_state,
                "slots": [],
                "migrations": [],
            }
            if len(parts) >= 9:
                slots, migrations = parse_slots(parts[8])
                node["slots"], node["migrations"] = slots, migrations

            nodes.append(node)

        return nodes


class ClusterShardsCallback(
    ResponseCallback[
        list[dict[AnyStr, list[RedisValueT] | Mapping[AnyStr, RedisValueT]]],
        list[dict[AnyStr, list[RedisValueT] | Mapping[AnyStr, RedisValueT]]],
    ]
):
    def transform(
        self,
        response: list[dict[AnyStr, list[RedisValueT] | Mapping[AnyStr, RedisValueT]]],
    ) -> list[dict[AnyStr, list[RedisValueT] | Mapping[AnyStr, RedisValueT]]]:
        return response



"""
RESP protocol constants
"""

from __future__ import annotations

import enum
from typing import Final

from coredis._utils import b


class DataType(enum.IntEnum):
    """
    Markers used by redis server to signal
    the type of data being sent.

    See:

    - `RESP protocol spec <https://redis.io/docs/develop/reference/protocol-spec>`__
    - `RESP3 specification <https://github.com/antirez/RESP3/blob/master/spec.md>`__
    """

    NONE = ord(b"_")
    SIMPLE_STRING = ord(b"+")
    BULK_STRING = ord(b"$")
    VERBATIM = ord(b"=")
    BOOLEAN = ord(b"#")
    INT = ord(b":")
    DOUBLE = ord(b",")
    BIGNUMBER = ord(b"(")
    ARRAY = ord(b"*")
    PUSH = ord(b">")
    MAP = ord(b"%")
    SET = ord(b"~")
    ERROR = ord(b"-")
    ATTRIBUTE = ord(b"|")


SYM_STAR: Final[bytes] = b("*")
SYM_DOLLAR: Final[bytes] = b("$")
SYM_CRLF: Final[bytes] = b("\r\n")
SYM_LF: Final[bytes] = b("\n")
SYM_EMPTY: Final[bytes] = b("")
SYM_TRUE: Final[bytes] = b("t")



from __future__ import annotations

from typing import Any, cast

from coredis.response._callbacks import ResponseCallback
from coredis.response.types import GeoCoordinates, GeoSearchResult
from coredis.typing import AnyStr, Generic, ResponsePrimitive, StringT


class GeoSearchCallback(
    Generic[AnyStr],
    ResponseCallback[
        list[StringT | list[ResponsePrimitive | list[ResponsePrimitive]]],
        tuple[AnyStr | GeoSearchResult, ...],
    ],
):
    def transform(
        self,
        response: list[StringT | list[ResponsePrimitive | list[ResponsePrimitive]]],
        **options: Any,
    ) -> tuple[AnyStr | GeoSearchResult, ...]:
        if not (
            self.options.get("withdist")
            or self.options.get("withcoord")
            or self.options.get("withhash")
        ):
            return tuple(cast(list[AnyStr], response))

        results: list[GeoSearchResult] = []

        for result in response:
            chunk = cast(list[list[ResponsePrimitive | list[ResponsePrimitive]]], result)
            results.append(
                GeoSearchResult(
                    cast(StringT, chunk.pop(0)),
                    float(cast(StringT, chunk.pop(0))) if self.options.get("withdist") else None,
                    cast(int, chunk.pop(0)) if self.options.get("withhash") else None,
                    (
                        GeoCoordinates(*cast(list[float], chunk.pop(0)))
                        if self.options.get("withcoord")
                        else None
                    ),
                )
            )

        return tuple(results)


class GeoCoordinatesCallback(
    ResponseCallback[list[list[float] | None], tuple[GeoCoordinates | None, ...]]
):
    def transform(
        self, response: list[list[float] | None], **options: Any
    ) -> tuple[GeoCoordinates | None, ...]:
        return tuple(
            map(
                lambda ll: GeoCoordinates(float(ll[0]), float(ll[1])) if ll is not None else None,
                response,
            )
        )



from __future__ import annotations

from ._base import BaseConnection, BaseConnectionParams, Location
from ._cluster import ClusterConnection
from ._sentinel import SentinelManagedConnection
from ._tcp import TCPConnection, TCPLocation
from ._uds import UnixDomainSocketConnection, UnixDomainSocketLocation

#: For backward compatibility
Connection = TCPConnection
__all__ = [
    "BaseConnectionParams",
    "BaseConnection",
    "Connection",
    "TCPConnection",
    "Location",
    "TCPLocation",
    "SentinelManagedConnection",
    "UnixDomainSocketLocation",
    "UnixDomainSocketConnection",
    "ClusterConnection",
]



from __future__ import annotations

import contextlib
import contextvars
import functools
import inspect
import random
import textwrap
from abc import ABCMeta
from ssl import SSLContext
from typing import TYPE_CHECKING, Any, cast, overload

from anyio import sleep
from deprecated.sphinx import versionadded, versionchanged

from coredis._concurrency import gather
from coredis.client.basic import Client, Redis
from coredis.cluster._node import ClusterNodeLocation
from coredis.commands._key_spec import KeySpec
from coredis.commands._routing import NodeExecution
from coredis.commands._validators import mutually_inclusive_parameters
from coredis.commands.constants import CommandName
from coredis.connection._base import RedisSSLContext
from coredis.connection._tcp import TCPLocation
from coredis.credentials import AbstractCredentialProvider
from coredis.exceptions import (
    AskError,
    ClusterDownError,
    ClusterError,
    ConnectionError,
    MovedError,
    RedisError,
    TryAgainError,
)
from coredis.globals import (
    CACHEABLE_COMMANDS,
    MODULE_GROUPS,
    READONLY_COMMANDS,
    ROUTING_STRATEGIES,
)
from coredis.patterns.cache import AbstractCache
from coredis.patterns.pubsub import ClusterPubSub, ShardedPubSub, SubscriptionCallback
from coredis.pool import ClusterConnectionPool
from coredis.response._callbacks import NoopCallback
from coredis.retry import (
    CompositeRetryPolicy,
    ConstantRetryPolicy,
    ExponentialBackoffRetryPolicy,
    RetryPolicy,
)
from coredis.typing import (
    AnyStr,
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    ExecutionParameters,
    Iterable,
    Iterator,
    KeyT,
    Literal,
    Mapping,
    Node,
    Parameters,
    ParamSpec,
    RedisCommand,
    RedisCommandP,
    Self,
    StringT,
    TypeAdapter,
    TypeVar,
    Unpack,
)

P = ParamSpec("P")
R = TypeVar("R")

if TYPE_CHECKING:
    import coredis.patterns.pipeline
    from coredis.patterns.lock import Lock
    from coredis.patterns.streams import Consumer, GroupConsumer, StreamParameters


class ClusterMeta(ABCMeta):
    def __new__(
        cls, name: str, bases: tuple[type, ...], namespace: dict[str, object]
    ) -> ClusterMeta:
        kls = super().__new__(cls, name, bases, namespace)
        methods = dict(k for k in inspect.getmembers(kls) if inspect.isfunction(k[1]))
        for module in MODULE_GROUPS:
            methods.update(
                {
                    f"{module.MODULE}.{k[0]}": k[1]
                    for k in inspect.getmembers(module)
                    if inspect.isfunction(k[1])
                }
            )
        for method_name, method in methods.items():
            doc_addition = ""
            cmd = getattr(method, "__coredis_command", None)
            if cmd:
                if not cmd.cluster.enabled:
                    doc_addition = """
.. warning:: Not supported in cluster mode
                    """
                else:
                    if cmd.cluster.routing_strategy:
                        doc_addition = f"""
.. admonition:: Cluster note

   The command will be **{cmd.cluster.routing_strategy.description}**
   and return {cmd.cluster.routing_strategy.merge_callback.response_policy}.

                    """
                        if cmd.cluster.routing_strategy.cross_slot:
                            doc_addition += """
   To disable this behavior set :paramref:`RedisCluster.non_atomic_cross_slot` to ``False``
                                            """
            if doc_addition and not hasattr(method, "__cluster_docs") and cmd:
                if not getattr(method, "__coredis_module", None):
                    if not cmd.cluster.enabled:

                        def __w(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
                            @functools.wraps(func)
                            def _w(*a: P.args, **k: P.kwargs) -> Awaitable[R]:
                                raise NotImplementedError(
                                    f"{func.__name__} is disabled for cluster client"
                                )

                            _w.__doc__ = f"""{textwrap.dedent(method.__doc__ or "")}
    {doc_addition}
                            """
                            return _w

                        wrapped = __w(method)
                        setattr(wrapped, "__cluster_docs", doc_addition)
                        setattr(kls, method_name, wrapped)
                    else:
                        setattr(method, "__cluster_docs", doc_addition)
                        method.__doc__ = f"""{textwrap.dedent(method.__doc__ or "")}
    {doc_addition}
                        """
                else:
                    method.__doc__ = f"""{textwrap.dedent(method.__doc__ or "")}
{doc_addition}
                    """
                    setattr(method, "__cluster_docs", doc_addition)
        return kls


RedisClusterT = TypeVar("RedisClusterT", bound="RedisCluster[Any]")


class RedisCluster(
    Client[AnyStr],
    metaclass=ClusterMeta,
):
    MAX_RETRIES = 16

    connection_pool: ClusterConnectionPool

    @overload
    def __init__(
        self: RedisCluster[bytes],
        host: str | None = ...,
        port: int | None = ...,
        *,
        startup_nodes: Iterable[Node] | None = ...,
        username: str | None = ...,
        password: str | None = ...,
        credential_provider: AbstractCredentialProvider | None = ...,
        stream_timeout: float | None = ...,
        connect_timeout: float | None = ...,
        pool_timeout: float | None = ...,
        ssl: bool = ...,
        ssl_context: SSLContext | None = ...,
        ssl_keyfile: str | None = ...,
        ssl_certfile: str | None = ...,
        ssl_cert_reqs: Literal["optional", "required", "none"] | None = ...,
        ssl_check_hostname: bool | None = ...,
        ssl_ca_certs: str | None = ...,
        max_connections: int = ...,
        max_connections_per_node: bool = ...,
        readonly: bool = ...,
        read_from_replicas: bool = ...,
        reinitialize_steps: int | None = ...,
        skip_full_coverage_check: bool = ...,
        nodemanager_follow_cluster: bool = ...,
        encoding: str = ...,
        decode_responses: Literal[False] = ...,
        connection_pool: ClusterConnectionPool | None = ...,
        connection_pool_cls: type[ClusterConnectionPool] = ...,
        verify_version: bool = ...,
        non_atomic_cross_slot: bool = ...,
        cache: AbstractCache | None = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        type_adapter: TypeAdapter | None = ...,
    ) -> None: ...

    @overload
    def __init__(
        self: RedisCluster[str],
        host: str | None = ...,
        port: int | None = ...,
        *,
        startup_nodes: Iterable[Node | TCPLocation] | None = ...,
        username: str | None = ...,
        password: str | None = ...,
        credential_provider: AbstractCredentialProvider | None = ...,
        stream_timeout: float | None = ...,
        connect_timeout: float | None = ...,
        pool_timeout: float | None = ...,
        ssl: bool = ...,
        ssl_context: SSLContext | None = ...,
        ssl_keyfile: str | None = ...,
        ssl_certfile: str | None = ...,
        ssl_cert_reqs: Literal["optional", "required", "none"] | None = ...,
        ssl_check_hostname: bool | None = ...,
        ssl_ca_certs: str | None = ...,
        max_connections: int = ...,
        max_connections_per_node: bool = ...,
        readonly: bool = ...,
        read_from_replicas: bool = ...,
        reinitialize_steps: int | None = ...,
        skip_full_coverage_check: bool = ...,
        nodemanager_follow_cluster: bool = ...,
        encoding: str = ...,
        decode_responses: Literal[True] = ...,
        connection_pool: ClusterConnectionPool | None = ...,
        connection_pool_cls: type[ClusterConnectionPool] = ...,
        verify_version: bool = ...,
        non_atomic_cross_slot: bool = ...,
        cache: AbstractCache | None = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        type_adapter: TypeAdapter | None = ...,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        host: str | None = ...,
        port: int | None = ...,
        *,
        startup_nodes: Iterable[Node | TCPLocation] | None = ...,
        username: str | None = ...,
        password: str | None = ...,
        credential_provider: AbstractCredentialProvider | None = ...,
        stream_timeout: float | None = ...,
        connect_timeout: float | None = ...,
        pool_timeout: float | None = ...,
        ssl: bool = ...,
        ssl_context: SSLContext | None = ...,
        ssl_keyfile: str | None = ...,
        ssl_certfile: str | None = ...,
        ssl_cert_reqs: Literal["optional", "required", "none"] | None = ...,
        ssl_check_hostname: bool | None = ...,
        ssl_ca_certs: str | None = ...,
        max_connections: int = ...,
        max_connections_per_node: bool = ...,
        readonly: bool = ...,
        read_from_replicas: bool = ...,
        reinitialize_steps: int | None = ...,
        skip_full_coverage_check: bool = ...,
        nodemanager_follow_cluster: bool = ...,
        encoding: str = ...,
        decode_responses: bool = ...,
        connection_pool: ClusterConnectionPool | None = ...,
        connection_pool_cls: type[ClusterConnectionPool] = ...,
        verify_version: bool = ...,
        non_atomic_cross_slot: bool = ...,
        cache: AbstractCache | None = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        type_adapter: TypeAdapter | None = ...,
    ) -> None: ...

    def __init__(
        self,
        host: str | None = None,
        port: int | None = None,
        *,
        startup_nodes: Iterable[Node | TCPLocation] | None = None,
        username: str | None = None,
        password: str | None = None,
        credential_provider: AbstractCredentialProvider | None = None,
        stream_timeout: float | None = None,
        connect_timeout: float | None = None,
        pool_timeout: float | None = None,
        ssl: bool = False,
        ssl_context: SSLContext | None = None,
        ssl_keyfile: str | None = None,
        ssl_certfile: str | None = None,
        ssl_cert_reqs: Literal["optional", "required", "none"] | None = None,
        ssl_check_hostname: bool | None = None,
        ssl_ca_certs: str | None = None,
        max_connections: int = 32,
        max_connections_per_node: bool = False,
        readonly: bool = False,
        read_from_replicas: bool = False,
        reinitialize_steps: int | None = None,
        skip_full_coverage_check: bool = False,
        nodemanager_follow_cluster: bool = True,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        connection_pool: ClusterConnectionPool | None = None,
        connection_pool_cls: type[ClusterConnectionPool] = ClusterConnectionPool,
        verify_version: bool = True,
        non_atomic_cross_slot: bool = True,
        cache: AbstractCache | None = None,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
        retry_policy: RetryPolicy = CompositeRetryPolicy(
            ConstantRetryPolicy((ClusterDownError,), retries=2, delay=0.1),
            ConstantRetryPolicy(
                (
                    ConnectionError,
                    TimeoutError,
                ),
                retries=2,
                delay=0.1,
            ),
        ),
        type_adapter: TypeAdapter | None = None,
    ) -> None:
        """

        Changes
          - .. versionremoved:: 6.0.0

            - :paramref:`protocol_version` removed (and therefore support for RESP2)

          - .. versionchanged:: 6.0.0

            -  The cluster client is now an async context manager and must always be used as such.

          - .. versionadded:: 4.12.0

            - :paramref:`retry_policy`
            - :paramref:`noevict`
            - :paramref:`notouch`
            - :meth:`RedisCluster.ensure_persistence` context manager
            - Redis Module support

              - RedisJSON: :attr:`RedisCluster.json`
              - RedisBloom:

                - BloomFilter: :attr:`RedisCluster.bf`
                - CuckooFilter: :attr:`RedisCluster.cf`
                - CountMinSketch: :attr:`RedisCluster.cms`
                - TopK: :attr:`RedisCluster.topk`
                - TDigest: :attr:`RedisCluster.tdigest`
              - RedisTimeSeries: :attr:`RedisCluster.timeseries`
              - RedisGraph: :attr:`RedisCluster.graph`
              - RediSearch:

                - Search & Aggregation: :attr:`RedisCluster.search`
                - Autocomplete: Added :attr:`RedisCluster.autocomplete`

          - .. versionchanged:: 4.4.0

            - :paramref:`nodemanager_follow_cluster` now defaults to ``True``

          - .. deprecated:: 4.4.0

            - The :paramref:`readonly` argument is deprecated in favour of
              :paramref:`read_from_replicas`

          - .. versionadded:: 4.3.0

            - Added :paramref:`connection_pool_cls`

          - .. versionchanged:: 4.0.0

            - :paramref:`non_atomic_cross_slot` defaults to ``True``
            - :paramref:`protocol_version`` defaults to ``3``

          - .. versionadded:: 3.11.0

            - Added :paramref:`noreply`

          - .. versionadded:: 3.10.0

            - Synchronized ssl constructor parameters with :class:`coredis.Redis`

          - .. versionadded:: 3.9.0

            - If :paramref:`cache` is provided the client will check & populate
              the cache for read only commands and invalidate it for commands
              that could change the key(s) in the request.

          - .. versionadded:: 3.6.0

            - The :paramref:`non_atomic_cross_slot` parameter was added

          - .. versionchanged:: 3.5.0

            - The :paramref:`verify_version` parameter now defaults to ``True``

          - .. versionadded:: 3.1.0

            - The :paramref:`protocol_version` and :paramref:`verify_version`
              parameters were added

        :param host: Can be used to point to a startup node
        :param port: Can be used to point to a startup node
        :param startup_nodes: List of nodes that initial bootstrapping can be done
         from
        :param stream_timeout: Timeout (seconds) when reading responses from the server
        :param connect_timeout: Timeout (seconds) for establishing a connection to the server
        :param pool_timeout: Timeout (seconds) for acquiring a connection from the
         connection pool
        :param ssl: Whether to use an SSL connection
        :param ssl_context: If provided the :class:`ssl.SSLContext` will be used when
         establishing the connection. Otherwise either the default context (if no other
         ssl related parameters are provided) or a custom context based on the other
         ``ssl_*`` parameters will be used.
        :param ssl_keyfile: Path of the private key to use
        :param ssl_certfile: Path to the certificate corresponding to :paramref:`ssl_keyfile`
        :param ssl_cert_reqs: Whether to try to verify the server's certificates and
         how to behave if verification fails (See :attr:`ssl.SSLContext.verify_mode`).
        :param ssl_check_hostname: Whether to enable hostname checking when establishing
         an ssl connection.
        :param ssl_ca_certs: Path to a concatenated certificate authority file or a directory
         containing several CA certifcates to use  for validating the server's certificates
         when :paramref:`ssl_cert_reqs` is not ``"none"``
         (See :meth:`ssl.SSLContext.load_verify_locations`).
        :param max_connections: Maximum number of connections that should be kept open at one time
        :param max_connections_per_node:
        :param read_from_replicas: If ``True`` the client will route readonly commands to replicas
        :param reinitialize_steps: Number of moved errors that result in a cluster
         topology refresh using the startup nodes provided
        :param skip_full_coverage_check: Skips the check of cluster-require-full-coverage config,
         useful for clusters without the CONFIG command (like aws)
        :param nodemanager_follow_cluster: The node manager will during initialization try the
         last set of nodes that it was operating on. This will allow the client to drift along
         side the cluster if the cluster nodes move around alot.
        :param encoding: The codec to use to encode strings transmitted to redis
         and decode responses with. (See :ref:`handbook/encoding:encoding/decoding`)
        :param decode_responses: If ``True`` string responses from the server
         will be decoded using :paramref:`encoding` before being returned.
         (See :ref:`handbook/encoding:encoding/decoding`)
        :param connection_pool: The connection pool instance to use. If not provided
         a new pool will be assigned to this client.
        :param connection_pool_cls: The connection pool class to use when constructing
         a connection pool for this instance.
        :param verify_version: Validate redis server version against the documented
         version introduced before executing a command and raises a
         :exc:`CommandNotSupportedError` error if the required version is higher than
         the reported server version
        :param non_atomic_cross_slot: If ``True`` certain commands that can operate
         on multiple keys (cross slot) will be split across the relevant nodes by
         mapping the keys to the appropriate slot and the result merged before being
         returned.
        :param cache: If provided the cache will be used to avoid requests for read only
         commands if the client has already requested the data and it hasn't been invalidated.
         The cache is responsible for any mutations to the keys that happen outside of this client
        :param noreply: If ``True`` the client will not request a response for any
         commands sent to the server.
        :param noevict: Ensures that connections from the client will be excluded from the
         client eviction process even if we're above the configured client eviction threshold.
        :param notouch: Ensures that commands sent by the client will not alter the LRU/LFU
         of the keys they access.
        :param retry_policy: The retry policy to use when interacting with the cluster
        :param type_adapter: The adapter to use for serializing / deserializing customs types
         when interacting with redis commands.
        """

        if connection_pool and cache:
            raise RuntimeError("Parameters 'cache' and 'connection_pool' are mutually exclusive!")
        if connection_pool:
            pool = connection_pool
        else:
            startup_nodes = list(
                node if isinstance(node, TCPLocation) else TCPLocation(node["host"], node["port"])
                for node in startup_nodes or []
            )

            # Support host/port as argument
            if host and not startup_nodes:
                startup_nodes.append(
                    TCPLocation(
                        host=host,
                        port=port if port else 7000,
                    )
                )
            if ssl_context is None and ssl:
                ssl_context = RedisSSLContext(
                    ssl_keyfile,
                    ssl_certfile,
                    ssl_cert_reqs,
                    ssl_ca_certs,
                    ssl_check_hostname,
                ).get()

            pool = connection_pool_cls(
                startup_nodes=startup_nodes,
                username=username,
                password=password,
                credential_provider=credential_provider,
                max_connections=max_connections,
                reinitialize_steps=reinitialize_steps,
                max_connections_per_node=max_connections_per_node,
                skip_full_coverage_check=skip_full_coverage_check,
                nodemanager_follow_cluster=nodemanager_follow_cluster,
                read_from_replicas=readonly or read_from_replicas,
                encoding=encoding,
                decode_responses=decode_responses,
                noreply=noreply,
                noevict=noevict,
                notouch=notouch,
                stream_timeout=stream_timeout,
                connect_timeout=connect_timeout,
                timeout=pool_timeout,
                ssl_context=ssl_context,
                _cache=cache,
            )

        super().__init__(
            connection_pool=pool,
            verify_version=verify_version,
            noreply=noreply,
            retry_policy=retry_policy,
            type_adapter=type_adapter,
        )

        self.non_atomic_cross_slot = non_atomic_cross_slot
        self._decodecontext: contextvars.ContextVar[bool | None,] = contextvars.ContextVar(
            "decode", default=None
        )
        self._encodingcontext: contextvars.ContextVar[str | None,] = contextvars.ContextVar(
            "decode", default=None
        )

    @classmethod
    @overload
    def from_url(
        cls,
        url: str,
        *,
        skip_full_coverage_check: bool = ...,
        decode_responses: Literal[False] = ...,
        verify_version: bool = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        non_atomic_cross_slot: bool = ...,
        cache: AbstractCache | None = ...,
        retry_policy: RetryPolicy = ...,
        type_adapter: TypeAdapter | None = ...,
        **kwargs: Any,
    ) -> RedisCluster[bytes]: ...

    @classmethod
    @overload
    def from_url(
        cls,
        url: str,
        *,
        skip_full_coverage_check: bool = ...,
        decode_responses: Literal[True],
        verify_version: bool = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        non_atomic_cross_slot: bool = ...,
        cache: AbstractCache | None = ...,
        retry_policy: RetryPolicy = ...,
        type_adapter: TypeAdapter | None = ...,
        **kwargs: Any,
    ) -> RedisCluster[str]: ...

    @classmethod
    def from_url(
        cls: type[RedisClusterT],
        url: str,
        *,
        skip_full_coverage_check: bool = False,
        decode_responses: bool = False,
        verify_version: bool = True,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
        non_atomic_cross_slot: bool = True,
        cache: AbstractCache | None = None,
        retry_policy: RetryPolicy = CompositeRetryPolicy(
            ConstantRetryPolicy((ClusterDownError,), retries=2, delay=0.1),
            ConstantRetryPolicy(
                (
                    ConnectionError,
                    TimeoutError,
                ),
                retries=2,
                delay=0.1,
            ),
        ),
        type_adapter: TypeAdapter | None = None,
        **kwargs: Any,
    ) -> RedisClusterT:
        """
        Return a Cluster client object configured from the startup node in URL,
        which must use either the ``redis://`` scheme
        `<http://www.iana.org/assignments/uri-schemes/prov/redis>`_

        For example:

            - ``redis://[:password]@localhost:6379``
            - ``rediss://[:password]@localhost:6379``

        :paramref:`url` and :paramref:`kwargs` are passed as is to
        the :func:`coredis.pool.ClusterConnectionPool.from_url`.
        """
        if decode_responses:
            return cls(
                decode_responses=True,
                verify_version=verify_version,
                noreply=noreply,
                retry_policy=retry_policy,
                type_adapter=type_adapter,
                connection_pool=ClusterConnectionPool.from_url(
                    url,
                    skip_full_coverage_check=skip_full_coverage_check,
                    decode_responses=decode_responses,
                    noreply=noreply,
                    noevict=noevict,
                    notouch=notouch,
                    _cache=cache,
                    **kwargs,
                ),
            )
        else:
            return cls(
                decode_responses=False,
                verify_version=verify_version,
                noreply=noreply,
                retry_policy=retry_policy,
                type_adapter=type_adapter,
                connection_pool=ClusterConnectionPool.from_url(
                    url,
                    skip_full_coverage_check=skip_full_coverage_check,
                    decode_responses=decode_responses,
                    noreply=noreply,
                    noevict=noevict,
                    notouch=notouch,
                    _cache=cache,
                    **kwargs,
                ),
            )

    @contextlib.asynccontextmanager
    async def __asynccontextmanager__(self) -> AsyncGenerator[Self]:
        async with self.connection_pool:
            yield self

    def __repr__(self) -> str:
        servers = list(node.name for node in self.connection_pool.cluster_layout.nodes)
        servers.sort()

        return "{}<{}>".format(type(self).__name__, ", ".join(servers))

    @property
    def all_nodes(self) -> Iterator[Redis[AnyStr]]:
        """ """
        for node in self.connection_pool.cluster_layout.nodes:
            yield node.as_client(**self.connection_pool.connection_kwargs)

    @property
    def primaries(self) -> Iterator[Redis[AnyStr]]:
        """ """
        for primary in self.connection_pool.cluster_layout.primaries:
            yield primary.as_client(**self.connection_pool.connection_kwargs)

    @property
    def replicas(self) -> Iterator[Redis[AnyStr]]:
        """ """
        for replica in self.connection_pool.cluster_layout.replicas:
            yield replica.as_client(**self.connection_pool.connection_kwargs)

    @property
    def num_replicas_per_shard(self) -> int:
        """
        Number of replicas per shard of the cluster determined by
        initial cluster topology discovery
        """
        if replicas := list(self.replicas):
            return int((len(list(self.all_nodes)) / len(replicas)) - 1)
        return 0

    async def on_retry_error(self, error: Exception) -> None:
        self.connection_pool.cluster_layout.report_errors(None, error)

    async def execute_command(
        self,
        command: RedisCommandP,
        callback: Callable[..., R] = NoopCallback(),
        **kwargs: Unpack[ExecutionParameters],
    ) -> R:
        """
        Sends a command to one or many nodes in the cluster
        with retries based on :paramref:`RedisCluster.retry_policy`
        """

        return await self.retry_policy.call_with_retries(
            lambda: self._execute_command(command, callback=callback, **kwargs),
            failure_hook=self.on_retry_error,
        )

    async def _execute_command(
        self,
        command: RedisCommandP,
        callback: Callable[..., R] = NoopCallback(),
        **kwargs: Unpack[ExecutionParameters],
    ) -> R:
        """
        Sends a command to one or many nodes in the cluster
        """
        prefer_replica = (
            command.name in READONLY_COMMANDS and self.connection_pool.read_from_replicas
        )
        affected_slots = len(
            KeySpec.affected_slots(
                command.name, *command.arguments, readonly_command=prefer_replica
            )
        )
        if affected_slots == 1:
            node = self.connection_pool.cluster_layout.node_for_request(
                command.name,
                command.arguments,
                primary=not prefer_replica,
                execution_parameters=kwargs,
            )
            return await self._execute_command_on_single_node(
                node,
                command,
                callback=callback,
                **kwargs,
            )
        elif (routing_strategy := ROUTING_STRATEGIES.get(command.name)) and (
            self.non_atomic_cross_slot or not routing_strategy.cross_slot
        ):
            node_executions = routing_strategy.distribute(
                self.connection_pool.cluster_layout,
                command.name,
                command.arguments,
                prefer_replica,
                execution_parameters=kwargs,
            )
            node_command_executions: dict[NodeExecution[R], Coroutine[Any, Any, R]] = {}
            for portion, execution in enumerate(node_executions):
                node_command_executions[execution] = self._execute_command_on_single_node(
                    execution.node,
                    RedisCommand(command.name, execution.arguments),
                    callback=callback,
                    **kwargs,
                )

            results = await gather(*node_command_executions.values(), return_exceptions=True)
            if self.noreply:
                return None  # type: ignore
            else:
                for execution, result in zip(node_command_executions.keys(), results):
                    execution.result = result
                return cast(
                    R,
                    routing_strategy.combine(node_executions=node_executions),
                )
        else:
            return await self._execute_command_on_single_node(
                self.connection_pool.cluster_layout.random_node(not prefer_replica),
                command,
                callback=callback,
                **kwargs,
            )

    async def _execute_command_on_single_node(
        self,
        node: ClusterNodeLocation,
        command: RedisCommandP,
        callback: Callable[..., R] = NoopCallback(),
        **kwargs: Unpack[ExecutionParameters],
    ) -> R:
        redirect_location = None
        asking = False

        remaining_attempts = int(self.MAX_RETRIES)
        quick_release = self.should_quick_release(command)
        should_block = not quick_release or self.requires_wait or self.requires_waitaof
        attempt_node: ClusterNodeLocation | None = node
        while remaining_attempts > 0:
            remaining_attempts -= 1
            released = False
            if redirect_location:
                attempt_node = self.connection_pool.cluster_layout.node_for_location(
                    redirect_location
                )
                redirect_location = None
            r = await self.connection_pool.get_connection(attempt_node, primary=node.is_primary)
            try:
                if asking:
                    await r.create_request(CommandName.ASKING, noreply=self.noreply, decode=False)
                    asking = False
                keys = KeySpec.extract_keys(command.name, *command.arguments)[0]
                cacheable = (
                    self.connection_pool.cache
                    and command.name in CACHEABLE_COMMANDS
                    and len(keys) == 1
                    and not self.noreply
                    and self._decodecontext.get() is None
                )
                cache_hit = False
                cached_reply = None
                use_cached = False
                reply = None
                if self.connection_pool.cache and self.connection_pool.cache.healthy:
                    if r.tracking_client_id != self.connection_pool.cache.get_client_id(r):
                        self.connection_pool.cache.reset()
                        await r.update_tracking_client(
                            True, self.connection_pool.cache.get_client_id(r)
                        )
                    if command.name not in READONLY_COMMANDS:
                        self.connection_pool.cache.invalidate(*keys)
                    elif cacheable:
                        try:
                            cached_reply = cast(
                                R,
                                self.connection_pool.cache.get(
                                    command.name,
                                    keys[0],
                                    *command.arguments,
                                ),
                            )
                            use_cached = random.random() * 100.0 < min(
                                100.0, self.connection_pool.cache.confidence
                            )
                            cache_hit = True
                        except KeyError:
                            pass

                if not (use_cached and cached_reply):
                    request = r.create_request(
                        command.name,
                        *command.arguments,
                        noreply=self.noreply,
                        decode=kwargs.get("decode", self._decodecontext.get()),
                        encoding=self._encodingcontext.get(),
                        disconnect_on_cancellation=should_block,
                    )
                    # TODO: Fix this! using both the release & should_block
                    #  flags to decide release logic is fragile. We should be
                    #  releasing early even in the cached response flow.
                    if not should_block:
                        released = True
                        self.connection_pool.release(r)

                    reply = await request
                    await self._ensure_wait_and_persist(command, r)
                if self.noreply:
                    return  # type: ignore
                else:
                    response = callback(
                        cached_reply if cache_hit else reply,
                    )
                    if self.connection_pool.cache and cacheable:
                        if cache_hit and not use_cached:
                            self.connection_pool.cache.feedback(
                                command.name,
                                keys[0],
                                *command.arguments,
                                match=cached_reply == reply,
                            )
                        if not cache_hit:
                            self.connection_pool.cache.put(
                                command.name,
                                keys[0],
                                *command.arguments,
                                value=reply,
                            )
                    return response
            except MovedError as e:
                self.connection_pool.cluster_layout.report_errors(attempt_node, e)
                redirect_location = TCPLocation(e.host, e.port)
                attempt_node = None
            except TryAgainError:
                if remaining_attempts < self.MAX_RETRIES / 2:
                    await sleep(0.05)
            except AskError as e:
                redirect_location, asking = TCPLocation(e.host, e.port), True
            except RedisError as err:
                self.connection_pool.cluster_layout.report_errors(attempt_node, err)
                raise
            finally:
                if r and not released:
                    self.connection_pool.release(r)
                self._ensure_server_version(r.server_version)

        raise ClusterError("Maximum retries exhausted.")

    @overload
    def decoding(
        self, mode: Literal[False], encoding: str | None = None
    ) -> contextlib.AbstractContextManager[RedisCluster[bytes]]: ...

    @overload
    def decoding(
        self, mode: Literal[True], encoding: str | None = None
    ) -> contextlib.AbstractContextManager[RedisCluster[str]]: ...

    @contextlib.contextmanager
    @versionadded(version="4.8.0")
    def decoding(self, mode: bool, encoding: str | None = None) -> Iterator[RedisCluster[Any]]:
        """
        Context manager to temporarily change the decoding behavior
        of the client

        :param mode: Whether to decode or not
        :param encoding: Optional encoding to use if decoding. If not provided
         the :paramref:`~coredis.RedisCluster.encoding` parameter provided to the client will
         be used.

        Example::

            client = coredis.RedisCluster(decode_responses=True)
            await client.set("fubar", "baz")
            assert await client.get("fubar") == "baz"
            with client.decoding(False):
                assert await client.get("fubar") == b"baz"
                with client.decoding(True):
                    assert await client.get("fubar") == "baz"

        """
        prev_decode = self._decodecontext.get()
        prev_encoding = self._encodingcontext.get()
        self._decodecontext.set(mode)
        self._encodingcontext.set(encoding)
        try:
            yield self
        finally:
            self._decodecontext.set(prev_decode)
            self._encodingcontext.set(prev_encoding)

    @versionchanged(version="6.0.0", reason="All arguments are now keyword only")
    def pubsub(
        self,
        *,
        channels: Parameters[StringT] | None = None,
        channel_handlers: Mapping[StringT, SubscriptionCallback] | None = None,
        patterns: Parameters[StringT] | None = None,
        pattern_handlers: Mapping[StringT, SubscriptionCallback] | None = None,
        ignore_subscribe_messages: bool = False,
        retry_policy: RetryPolicy | None = CompositeRetryPolicy(
            ExponentialBackoffRetryPolicy(
                (ConnectionError,), retries=None, base_delay=0.1, max_delay=16, jitter=True
            ),
            ConstantRetryPolicy((TimeoutError,), retries=2, delay=0.1),
        ),
        subscription_timeout: float = 1,
        max_idle_seconds: float = 15,
    ) -> ClusterPubSub[AnyStr]:
        """
        Return a Pub/Sub instance that can be used to consume messages that get
        published to the subscribed channels or patterns.

        :param channels: channels that the constructed Pubsub instance should
         automatically subscribe to
        :param channel_handlers: Mapping of channels to automatically subscribe to
         and the associated handlers that will be invoked when a message is received
         on the specific channel.
        :param patterns: patterns that the constructed Pubsub instance should
         automatically subscribe to
        :param pattern_handlers: Mapping of patterns to automatically subscribe to
         and the associated handlers that will be invoked when a message is received
         on channel matching the pattern.
        :param ignore_subscribe_messages: Whether to skip subscription
         acknowledgement messages
        :param retry_policy: An explicit retry policy to use in the subscriber.
        :param subscription_timeout: Maximum amount of time in seconds to wait for
         acknowledgement of subscriptions.
        :param max_idle_seconds: Maximum duration (in seconds) to tolerate no
         messages from the cluster before performing a keepalive check with a
         ``PING``.
        """
        return ClusterPubSub[AnyStr](
            self.connection_pool,
            ignore_subscribe_messages=ignore_subscribe_messages,
            retry_policy=retry_policy,
            channels=channels,
            channel_handlers=channel_handlers,
            patterns=patterns,
            pattern_handlers=pattern_handlers,
            subscription_timeout=subscription_timeout,
            max_idle_seconds=max_idle_seconds,
        )

    @versionadded(version="3.6.0")
    @versionchanged(version="6.0.0", reason="All arguments are now keyword only")
    def sharded_pubsub(
        self,
        *,
        channels: Parameters[StringT] | None = None,
        channel_handlers: Mapping[StringT, SubscriptionCallback] | None = None,
        ignore_subscribe_messages: bool = False,
        read_from_replicas: bool = False,
        retry_policy: RetryPolicy | None = CompositeRetryPolicy(
            ExponentialBackoffRetryPolicy(
                (ConnectionError,), retries=None, base_delay=0.1, max_delay=16, jitter=True
            ),
            ConstantRetryPolicy((TimeoutError,), retries=2, delay=0.1),
        ),
        subscription_timeout: float = 1,
        max_idle_seconds: float = 15,
    ) -> ShardedPubSub[AnyStr]:
        """
        Return a Pub/Sub instance that can be used to consume messages from
        the subscribed channels in a redis cluster.

        The implementation returned differs from that returned by :meth:`pubsub`
        as it uses the Sharded Pub/Sub implementation which routes messages
        to cluster nodes using the same algorithm used to assign keys to slots.
        This effectively restricts the propagation of messages to be within the
        shard of a cluster hence affording horizontally scaling the use of Pub/Sub
        with the cluster itself.

        :param channels: channels that the constructed Pubsub instance should
         automatically subscribe to
        :param channel_handlers: Mapping of channels to automatically subscribe to
         and the associated handlers that will be invoked when a message is received
         on the specific channel.
        :param ignore_subscribe_messages: Whether to skip subscription
         acknowledgement messages
        :param read_from_replicas: Whether to read messages from replica nodes
        :param retry_policy: An explicit retry policy to use in the subscriber.
        :param subscription_timeout: Maximum amount of time in seconds to wait for
         acknowledgement of subscriptions.
        :param max_idle_seconds: Maximum duration (in seconds) to tolerate no
         messages from the cluster before performing a keepalive check with a
         ``PING``.

        New in :redis-version:`7.0.0`
        """

        return ShardedPubSub[AnyStr](
            self.connection_pool,
            ignore_subscribe_messages=ignore_subscribe_messages,
            read_from_replicas=read_from_replicas,
            retry_policy=retry_policy,
            channels=channels,
            channel_handlers=channel_handlers,
            subscription_timeout=subscription_timeout,
            max_idle_seconds=max_idle_seconds,
        )

    def pipeline(
        self,
        transaction: bool = False,
        *,
        raise_on_error: bool = True,
        timeout: float | None = None,
    ) -> coredis.patterns.pipeline.ClusterPipeline[AnyStr]:
        """
        Returns a new pipeline object that can queue multiple commands for
        batch execution. Pipelines in cluster mode only provide a subset of the
        functionality of pipelines in standalone mode.

        Specifically:

        - Each command in the pipeline should only access keys on the same node
        - Transactions are disabled by default and are only supported if all
          watched keys route to the same node as where the commands in the multi/exec
          part of the pipeline.

        :param transaction: indicates whether all commands should be executed atomically.
        :param raise_on_error: Whether to raise errors upon executing the pipeline.
         If set to `False` errors will be accumulated and retrievable from the individual
         commands that had errors.
        :param watches: If :paramref:`transaction` is True these keys are watched for external
         changes during the transaction.
        :param timeout: If specified this value will take precedence over
         :paramref:`RedisCluster.stream_timeout`

        """

        from coredis.patterns.pipeline import ClusterPipeline

        return ClusterPipeline[AnyStr](
            client=self,
            raise_on_error=raise_on_error,
            transaction=transaction,
            timeout=timeout,
        )

    def lock(
        self,
        name: StringT,
        timeout: float | None = None,
        sleep: float = 0.1,
        blocking: bool = True,
        blocking_timeout: float | None = None,
    ) -> Lock[AnyStr]:
        """
        Return a lock instance which can be used to guard resource access across
        multiple clients.

        :param name: key for the lock
        :param timeout: indicates a maximum life for the lock.
         By default, it will remain locked until :meth:`~coredis.patterns.lock.Lock.release`
         is called.

        :param sleep: indicates the amount of time to sleep per loop iteration
         when the lock is in blocking mode and another client is currently
         holding the lock.

        :param blocking: indicates whether calling :meth:`~coredis.patterns.lock.Lock.acquire` should block until
         the lock has been acquired or to fail immediately, causing :meth:`acquire`
         to return ``False`` and the lock not being acquired. Defaults to ``True``.

        :param blocking_timeout: indicates the maximum amount of time in seconds to
         spend trying to acquire the lock. A value of ``None`` indicates
         continue trying forever.
        """
        from coredis.patterns.lock import Lock

        return Lock(self, name, timeout, sleep, blocking, blocking_timeout)

    @overload
    def xconsumer(
        self,
        streams: Parameters[KeyT],
        *,
        buffer_size: int = ...,
        timeout: int | None = ...,
        group: Literal[None] = ...,
        consumer: Literal[None] = ...,
        auto_create: bool = ...,
        auto_acknowledge: bool = ...,
        start_from_backlog: bool = ...,
        **stream_parameters: StreamParameters,
    ) -> Consumer[AnyStr]: ...
    @overload
    def xconsumer(
        self,
        streams: Parameters[KeyT],
        *,
        buffer_size: int = ...,
        timeout: int | None = ...,
        group: StringT = ...,
        consumer: StringT = ...,
        auto_create: bool = ...,
        auto_acknowledge: bool = ...,
        start_from_backlog: bool = ...,
        **stream_parameters: StreamParameters,
    ) -> GroupConsumer[AnyStr]: ...
    @versionadded(version="6.0.0")
    @mutually_inclusive_parameters("group", "consumer")
    def xconsumer(
        self,
        streams: Parameters[KeyT],
        *,
        buffer_size: int = 0,
        timeout: int | None = None,
        group: StringT | None = None,
        consumer: StringT | None = None,
        auto_create: bool = True,
        auto_acknowledge: bool = False,
        start_from_backlog: bool = False,
        **stream_parameters: StreamParameters,
    ) -> Consumer[AnyStr] | GroupConsumer[AnyStr]:
        """
        Create a stream consumer for one or more Redis streams.

        Depending on whether ``group`` and ``consumer`` are provided, this method
        creates either a standalone stream consumer or a member of a stream
        consumer group.

        If ``group`` and ``consumer`` are not provided, a standalone stream
        consumer is created that starts reading from the latest entry of each
        stream provided in :paramref:`streams`.

        The latest entry is determined by calling
        :meth:`~coredis.RedisCluster.xinfo_stream` and using the :data:`last-entry`
        attribute at the point of initializing the consumer instance or on first
        fetch (whichever comes first). If the stream(s) do not exist at the time
        of consumer creation, the consumer will simply start from the minimum
        identifier (``0-0``).

        If ``group`` and ``consumer`` are provided, a member of a stream consumer
        group is created. The consumer has an identical interface as
        :class:`coredis.patterns.streams.Consumer`.

        :param streams: the stream identifiers to consume from
        :param buffer_size: Size of buffer (per stream) to maintain. This
         translates to the maximum number of stream entries that are fetched
         on each request to redis.
        :param timeout: Maximum amount of time in milliseconds to block for new
         entries to appear on the streams the consumer is reading from.
        :param group: The name of the group this consumer is part of
        :param consumer: The unique name (within :paramref:`group`) of the consumer
        :param auto_create: If True the group will be created upon initialization
         or first fetch if it doesn't already exist.
        :param auto_acknowledge: If ``True`` the stream entries fetched will be fetched
         without needing to be acknowledged with :meth:`~coredis.RedisCluster.xack` to remove
         them from the pending entries list.
        :param start_from_backlog: If ``True`` the consumer will start by fetching any pending
         entries from the pending entry list before considering any new messages
         not seen by any other consumer in the :paramref:`group`
        :param stream_parameters: Mapping of optional parameters to use
         by stream for the streams provided in :paramref:`streams`.

         .. warning:: Providing an ``identifier`` in ``stream_parameters`` has a different
            meaning for a group consumer. If the value is any valid identifier other than ``>``
            the consumer will only access the history of pending messages. That is, the set of
            messages that were delivered to this consumer (identified by :paramref:`consumer`)
            and never acknowledged.
        """
        from coredis.patterns.streams import Consumer, GroupConsumer

        if group is not None and consumer is not None:
            return GroupConsumer(
                self,
                streams,
                group=group,
                consumer=consumer,
                buffer_size=buffer_size,
                auto_create=auto_create,
                auto_acknowledge=auto_acknowledge,
                start_from_backlog=start_from_backlog,
                timeout=timeout,
                **stream_parameters,
            )
        else:
            return Consumer(
                self, streams, buffer_size=buffer_size, timeout=timeout, **stream_parameters
            )

    async def scan_iter(
        self,
        match: StringT | None = None,
        count: int | None = None,
        type_: StringT | None = None,
    ) -> AsyncIterator[AnyStr]:
        for node in self.primaries:
            async with node:
                cursor = None
                while cursor != 0:
                    cursor, data = await node.scan(cursor or 0, match, count, type_)
                    for item in data:
                        yield item



from __future__ import annotations

import dataclasses

from anyio import connect_unix, fail_after
from anyio.abc import ByteStream

from coredis.typing import Unpack

from ._base import BaseConnection, BaseConnectionParams, Location


@dataclasses.dataclass(unsafe_hash=True)
class UnixDomainSocketLocation(Location):
    """Location of a redis instance listening on a unix domain socket"""

    #: The absolute path of the socket
    path: str

    def __repr__(self) -> str:
        return f"<path={self.path}>"

    async def check(self) -> bool:
        try:
            async with await connect_unix(self.path):
                return True
        except OSError:
            return False
        return False


class UnixDomainSocketConnection(BaseConnection):
    location: UnixDomainSocketLocation

    def __init__(
        self,
        location: UnixDomainSocketLocation,
        **kwargs: Unpack[BaseConnectionParams],
    ):
        super().__init__(location, **kwargs)
        # FIXME: this is only for backward compatibility as 6.0 still had
        # path in uds connections
        self.path = self.location.path

    async def _connect(self) -> ByteStream:
        with fail_after(self._connect_timeout):
            return await connect_unix(self.location.path)

    def describe(self) -> str:
        return f"UnixDomainSocketConnection<path={self.location.path},db={self._db}>"



from __future__ import annotations

from collections.abc import Hashable
from typing import Any, TypeAlias

from .typing import MutableSet, RedisError, StringT

#: Primitives returned by redis
ResponsePrimitive: TypeAlias = StringT | int | float | bool | None

#: Represents the total structure of any response for any redis command.
ResponseType: TypeAlias = (
    ResponsePrimitive
    | list[Any]
    | MutableSet[Hashable]
    | dict[
        Hashable,
        Any,
    ]
    | RedisError
)
#: Type alias for valid python types that can be represented as json
JsonType: TypeAlias = str | int | float | bool | dict[str, Any] | list[Any] | None



from __future__ import annotations

import warnings

from coredis.patterns.streams import (
    Consumer,
    GroupConsumer,
    StreamParameters,
)

warnings.warn(
    "coredis.stream has been moved to coredis.patterns.streams",
    DeprecationWarning,
    stacklevel=2,
)
__all__ = ["Consumer", "GroupConsumer", "StreamParameters"]



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import Any, cast

from anyio import AsyncContextManagerMixin
from deprecated.sphinx import versionchanged

from coredis._utils import nativestr
from coredis.client import Client, RedisCluster
from coredis.cluster._node import ClusterNodeLocation
from coredis.commands import CommandRequest, CommandResponseT
from coredis.commands._key_spec import KeySpec
from coredis.commands.constants import CommandName
from coredis.commands.script import Script
from coredis.connection._base import BaseConnection, CommandInvocation
from coredis.connection._request import Request
from coredis.exceptions import (
    AskError,
    ClusterCrossSlotError,
    ClusterDownError,
    ClusterTransactionError,
    ConnectionError,
    ExecAbortError,
    MovedError,
    RedisClusterError,
    RedisError,
    ResponseError,
    TryAgainError,
    WatchError,
)
from coredis.pool import ClusterConnectionPool
from coredis.response._callbacks import (
    AnyStrCallback,
    BoolsCallback,
    NoopCallback,
)
from coredis.retry import ConstantRetryPolicy, retryable
from coredis.typing import (
    AnyStr,
    AsyncGenerator,
    Awaitable,
    Callable,
    ExecutionParameters,
    Generator,
    Iterable,
    KeyT,
    ParamSpec,
    RedisCommand,
    RedisCommandP,
    RedisValueT,
    ResponseType,
    Self,
    T_co,
    TypeVar,
    Unpack,
    ValueT,
)

P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")

ERRORS_ALLOW_RETRY = (
    MovedError,
    AskError,
    TryAgainError,
)

UNWATCH_COMMANDS = {CommandName.DISCARD, CommandName.EXEC, CommandName.UNWATCH}


class PipelineResult(Awaitable[T]):
    __slots__ = "_result"

    def __init__(self, result: T) -> None:
        self._result = result

    @property
    def value(self) -> T:
        if isinstance(self._result, Exception):
            raise self._result
        else:
            return self._result

    def __await__(self) -> Generator[Any, None, T]:
        async def _coro() -> T:
            return self.value

        return _coro().__await__()


class PipelineCommandRequest(CommandRequest[CommandResponseT]):
    """
    Command request returned by a pipeline command
    """

    client: Pipeline[Any]

    def __init__(
        self,
        client: Pipeline[Any],
        name: bytes,
        *arguments: ValueT,
        callback: Callable[..., CommandResponseT],
        execution_parameters: ExecutionParameters | None = None,
    ) -> None:
        super().__init__(
            client, name, *arguments, callback=callback, execution_parameters=execution_parameters
        )
        client._pipeline_execute_command(self)

    def __await__(self) -> Generator[None, None, CommandResponseT]:
        if hasattr(self, "_response"):
            return self._response.__await__()
        raise RuntimeError("You can't await a pipeline command before it completes executing")


class ClusterPipelineCommandRequest(CommandRequest[CommandResponseT]):
    """
    Command request for cluster pipelines, tracks position and result for cluster routing.
    """

    client: ClusterPipeline[Any]

    def __init__(
        self,
        client: ClusterPipeline[Any],
        name: bytes,
        *arguments: ValueT,
        callback: Callable[..., CommandResponseT],
        execution_parameters: ExecutionParameters | None = None,
    ) -> None:
        super().__init__(
            client, name, *arguments, callback=callback, execution_parameters=execution_parameters
        )
        self.position: int = 0
        self.result: Any = None
        self.asking: bool = False
        client._pipeline_execute_command(self)

    def __await__(self) -> Generator[None, None, CommandResponseT]:
        if hasattr(self, "_response"):
            return self._response.__await__()
        raise RuntimeError("You can't await a pipeline command before it completes executing")


class NodeCommands(AsyncContextManagerMixin):
    """
    Helper for grouping and executing commands on a single cluster node, handling transactions if needed.
    """

    connection: BaseConnection

    def __init__(
        self,
        client: RedisCluster[AnyStr],
        node: ClusterNodeLocation,
        connection: BaseConnection | None = None,
        in_transaction: bool = False,
        timeout: float | None = None,
        raise_on_error: bool = True,
    ):
        self.client: RedisCluster[Any] = client
        self.node = node
        self._connection = connection
        self.commands: list[ClusterPipelineCommandRequest[Any]] = []
        self.in_transaction = in_transaction
        self.timeout = timeout
        self._raise_on_error = raise_on_error
        self.multi_cmd: Request | None = None
        self.exec_cmd: Request | None = None

    def extend(self, c: list[ClusterPipelineCommandRequest[Any]]) -> None:
        self.commands.extend(c)

    def append(self, c: ClusterPipelineCommandRequest[Any]) -> None:
        self.commands.append(c)

    @asynccontextmanager
    async def __asynccontextmanager__(self) -> AsyncGenerator[None]:
        if not self._connection:
            async with self.client.connection_pool.acquire(node=self.node) as self.connection:
                yield
        else:
            self.connection = self._connection
            yield

    def write(self) -> None:
        connection = self.connection
        commands = self.commands

        # Reset results for all commands before writing.
        for c in commands:
            c.result = None

        # Batch all commands into a single request for efficiency.
        try:
            if self.in_transaction:
                self.multi_cmd = connection.create_request(CommandName.MULTI, timeout=self.timeout)
            requests = connection.create_requests(
                [
                    CommandInvocation(
                        cmd.name,
                        cmd.arguments,
                        (
                            bool(cmd.execution_parameters.get("decode"))
                            if cmd.execution_parameters.get("decode")
                            else None
                        ),
                        None,
                    )
                    for cmd in commands
                ],
                timeout=self.timeout,
            )
            if self.in_transaction:
                self.exec_cmd = connection.create_request(CommandName.EXEC, timeout=self.timeout)
            for i, cmd in enumerate(commands):
                cmd._response = requests[i]
        except (ConnectionError, TimeoutError) as e:
            for c in commands:
                c.result = e

    async def read(self) -> None:
        success = True
        multi_result = None
        if self.multi_cmd:
            multi_result = await self.multi_cmd
            success = multi_result in {b"OK", "OK"}
        for c in self.commands:
            if c.result is None:
                try:
                    c.result = await c._response if c._response else None
                except ExecAbortError:
                    raise
                except (ConnectionError, TimeoutError, RedisError) as e:
                    success = False
                    c.result = e
        if self.in_transaction and self.exec_cmd:
            if success:
                res = await self.exec_cmd
                if res:
                    transaction_result = cast(list[ResponseType], res)
                else:
                    raise WatchError("Watched variable changed.")
                for idx, c in enumerate(
                    [
                        _c
                        for _c in sorted(self.commands, key=lambda x: x.position)
                        if _c.name not in {CommandName.MULTI, CommandName.EXEC}
                    ]
                ):
                    c.result = c.callback(
                        transaction_result[idx],
                    )
                    c._response = PipelineResult(c.result)
            elif isinstance(multi_result, BaseException):
                raise multi_result


@versionchanged(
    version="6.0.0",
    reason="Pipelines are no longer awaitable. They support the async context manager protocol and must always be used as such",
)
class Pipeline(Client[AnyStr]):
    """
    Pipeline for batching multiple commands to a Redis server.
    Supports transactions and command stacking.

    All commands executed within a pipeline are wrapped with MULTI and EXEC
    calls when :paramref:`transaction` is ``True``.

    Any command raising an exception does **not** halt the execution of
    subsequent commands in the pipeline, however the first exception encountered
    will be raised when exiting the pipeline if :paramref:`raise_on_error` is ``True``.
    If not the exception is caught and will be returned when awaiting the command that failed.
    """

    QUEUED_RESPONSES = {b"QUEUED", "QUEUED"}

    def __init__(
        self,
        client: Client[AnyStr],
        transaction: bool | None,
        raise_on_error: bool = True,
        timeout: float | None = None,
    ) -> None:
        """
        :param transaction: Whether to wrap the commands in the pipeline in a ``MULTI``, ``EXEC``
        :param raise_on_error: Whether to raise the first error encounterd in the pipeline after
         executing it
        :param timeout: Time in seconds to wait for the pipeline results to return
        """
        self.client: Client[AnyStr] = client
        self._connection: BaseConnection | None = None
        self._transaction = transaction
        self._raise_on_error = raise_on_error
        self.watching = False
        self.command_stack: list[PipelineCommandRequest[Any]] = []
        self.watches: list[KeyT] = []
        self.cache = None
        self.explicit_transaction = False
        self.scripts: set[Script[AnyStr]] = set()
        self.timeout = timeout
        self.type_adapter = client.type_adapter
        self._results: tuple[Any] | None = None

    @asynccontextmanager
    async def __asynccontextmanager__(self) -> AsyncGenerator[Self]:
        yield self
        await self._execute()
        if self._connection:
            self.client.connection_pool.release(self._connection)

    def __repr__(self) -> str:
        return f"{type(self).__name__}<{repr(self._connection)}>"

    def create_request(
        self,
        name: bytes,
        *arguments: ValueT,
        callback: Callable[..., T_co],
        execution_parameters: ExecutionParameters | None = None,
    ) -> CommandRequest[T_co]:
        """
        :meta private:
        """
        return PipelineCommandRequest(
            self, name, *arguments, callback=callback, execution_parameters=execution_parameters
        )

    @asynccontextmanager
    async def watch(self, *keys: KeyT) -> AsyncGenerator[None]:
        """
        The given keys will be watched for changes within this context and the
        commands stacked within the context will be automatically executed when the
        context exits.
        """
        if self.command_stack:
            raise WatchError("Unable to add a watch after pipeline commands have been added")
        if not self._connection:
            self._connection = await self.client.connection_pool.get_connection()
        self.watches.extend(keys)
        await self._immediate_execute_command(
            RedisCommand(CommandName.WATCH, arguments=tuple(self.watches))
        )
        self.explicit_transaction = True
        yield
        await self._execute()

    @property
    def results(self) -> tuple[Any, ...] | None:
        """
        The results of the pipeline execution which can be accessed
        after the pipeline has completed.
        """
        if self.command_stack:
            raise RuntimeError("Pipeline results are not available before it completes execution")
        return self._results

    def execute_command(
        self,
        command: RedisCommandP,
        callback: Callable[..., R] = NoopCallback(),
        **options: Unpack[ExecutionParameters],
    ) -> Awaitable[R]:
        raise NotImplementedError

    async def _clear(self) -> None:
        """
        Clear the pipeline and reset state.
        """
        self.command_stack.clear()
        self.scripts.clear()
        # Reset connection state if we were watching something.
        if self.watches and self._connection:
            await self._connection.create_request(CommandName.UNWATCH, decode=False)
        self.watches.clear()
        self.explicit_transaction = False

    async def _immediate_execute_command(
        self,
        command: RedisCommandP,
        callback: Callable[..., R] = NoopCallback(),
        **kwargs: Unpack[ExecutionParameters],
    ) -> R:
        """
        Executes a command immediately, but don't auto-retry on a
        ConnectionError if we're already WATCHing a variable. Used when
        issuing WATCH or subsequent commands retrieving their values but before
        MULTI is called.

        :meta private:
        """
        assert self._connection
        request = self._connection.create_request(
            command.name, *command.arguments, decode=kwargs.get("decode")
        )
        return callback(await request)

    def _pipeline_execute_command(
        self,
        command: PipelineCommandRequest[R],
    ) -> None:
        """
        Queue a command for execution

        :meta private:
        """
        self.command_stack.append(command)

    async def _execute_transaction(
        self,
        connection: BaseConnection,
        commands: list[PipelineCommandRequest[Any]],
    ) -> None:
        requests = connection.create_requests(
            [CommandInvocation(CommandName.MULTI, (), None, None)]
            + [
                CommandInvocation(
                    cmd.name,
                    cmd.arguments,
                    (
                        bool(cmd.execution_parameters.get("decode"))
                        if cmd.execution_parameters.get("decode")
                        else None
                    ),
                    None,
                )
                for cmd in commands
            ]
            + [CommandInvocation(CommandName.EXEC, (), None, None)],
            timeout=self.timeout,
        )

        errors: list[tuple[int, RedisError | TimeoutError | None]] = []
        # parse off the response for MULTI
        # NOTE: we need to handle ResponseErrors here and continue
        # so that we read all the additional command messages from
        # the socket
        try:
            await requests[0]
        except (RedisError, TimeoutError) as e:
            errors.append((0, e))

        # and all the other commands
        for i, cmd in enumerate(commands):
            try:
                if (resp := await requests[i + 1]) not in self.QUEUED_RESPONSES:
                    raise Exception(
                        f"Abnormal response in pipeline for command {cmd.name!r}: {resp!r}"
                    )
            except (RedisError, TimeoutError) as e:
                self._annotate_exception(e, i + 1, cmd.name, cmd.arguments)
                errors.append((i + 1, e))

        try:
            response = cast(list[ResponseType] | None, await requests[-1])
        except (ExecAbortError, ResponseError, TimeoutError) as e:
            if errors and errors[0][1]:
                raise errors[0][1] from e
            raise

        if response is None:
            raise WatchError("Watched variable changed.")

        # put any parse errors into the response
        for i, e in errors:  # type: ignore
            response.insert(i, cast(ResponseType, e))

        if len(response) != len(commands):
            raise ResponseError("Wrong number of response items from pipeline execution")

        # We have to run response callbacks manually
        data: list[Any] = []
        for r, cmd in zip(response, commands):
            if not isinstance(r, Exception):
                r = cmd.callback(r, **cmd.execution_parameters)
            cmd._response = PipelineResult(r)
            data.append(r)

        self._results = tuple(data)
        # find any errors in the response and raise if necessary
        if self._raise_on_error:
            self._raise_first_error(commands, response)

    async def _execute_pipeline(
        self, connection: BaseConnection, commands: list[PipelineCommandRequest[Any]]
    ) -> None:
        # build up all commands into a single request to increase network perf
        requests = connection.create_requests(
            [
                CommandInvocation(
                    cmd.name,
                    cmd.arguments,
                    (
                        bool(cmd.execution_parameters.get("decode"))
                        if cmd.execution_parameters.get("decode")
                        else None
                    ),
                    None,
                )
                for cmd in commands
            ],
            timeout=self.timeout,
        )
        for i, cmd in enumerate(commands):
            cmd._response = requests[i]

        response: list[Any] = []
        for cmd in commands:
            try:
                res = await cmd._response if cmd._response else None
                resp = cmd.callback(
                    res,
                    **cmd.execution_parameters,
                )
                cmd._response = PipelineResult(resp)
                response.append(resp)
            except (ResponseError, TimeoutError) as re:
                cmd._response = PipelineResult(re)
                response.append(re)
        self._results = tuple(response)
        if self._raise_on_error:
            self._raise_first_error(commands, response)

    def _raise_first_error(
        self, commands: list[PipelineCommandRequest[Any]], response: ResponseType
    ) -> None:
        assert isinstance(response, list)
        for i, r in enumerate(response):
            if isinstance(r, (RedisError, TimeoutError)):
                self._annotate_exception(r, i + 1, commands[i].name, commands[i].arguments)
                raise r

    def _annotate_exception(
        self,
        exception: RedisError | TimeoutError | None,
        number: int,
        command: bytes,
        args: Iterable[RedisValueT],
    ) -> None:
        if exception:
            cmd = command.decode("latin-1")
            args = " ".join(map(str, args))
            msg = f"Command # {number} ({cmd} {args}) of pipeline caused error: {str(exception.args[0])}"
            exception.args = (msg,) + exception.args[1:]

    async def _load_scripts(self) -> None:
        # make sure all scripts that are about to be run on this pipeline exist
        scripts = list(self.scripts)
        shas = [s.sha for s in scripts]
        exists = await self._immediate_execute_command(
            RedisCommand(CommandName.SCRIPT_EXISTS, tuple(shas)), callback=BoolsCallback()
        )

        if not all(exists):
            for s, exist in zip(scripts, exists):
                if not exist:
                    s.sha = await self._immediate_execute_command(
                        RedisCommand(CommandName.SCRIPT_LOAD, (s.script,)),
                        callback=AnyStrCallback[AnyStr](),
                    )

    async def _execute(self) -> None:
        """
        Execute all queued commands in the pipeline.
        """
        if not self.command_stack:
            return None
        if not self._connection:
            self._connection = await self.client.connection_pool.get_connection()

        if self.scripts:
            await self._load_scripts()
        if self._transaction or self.explicit_transaction:
            exec = self._execute_transaction
        else:
            exec = self._execute_pipeline

        try:
            return await exec(self._connection, self.command_stack)
        except (ConnectionError, TimeoutError) as e:
            if self.watches:
                raise WatchError(
                    "A connection error occurred while watching one or more keys"
                ) from e
            raise
        finally:
            await self._clear()


@versionchanged(
    version="6.0.0",
    reason="Cluster Pipelines are no longer awaitable. They support the async context manager protocol and must always be used as such",
)
class ClusterPipeline(Client[AnyStr]):
    """
    Pipeline for batching multiple commands to a Redis Cluster
    Supports transactions only when all keys map to the same shard, and therefore
    :paramref:`transactions` is set to ``False`` by default due to the limited scope.

    Any command raising an exception does **not** halt the execution of
    subsequent commands in the pipeline, however the first exception encountered
    will be raised when exiting the pipeline if :paramref:`raise_on_error` is ``True``.
    If not the exception is caught and will be returned when awaiting the command that failed.
    """

    client: RedisCluster[AnyStr]
    connection_pool: ClusterConnectionPool
    command_stack: list[ClusterPipelineCommandRequest[Any]]

    def __init__(
        self,
        client: RedisCluster[AnyStr],
        raise_on_error: bool = True,
        transaction: bool = False,
        timeout: float | None = None,
    ) -> None:
        """
        :param transaction: Whether to wrap the commands in the pipeline in a ``MULTI``, ``EXEC``
        :param raise_on_error: Whether to raise the first error encounterd in the pipeline after
         executing it
        :param timeout: Time in seconds to wait for the pipeline results to return
        """
        self.command_stack = []
        self.client = client
        self.connection_pool = client.connection_pool
        self._raise_on_error = raise_on_error
        self._transaction = transaction
        self._watched_node: ClusterNodeLocation | None = None
        self._watched_connection: BaseConnection | None = None
        self.watches: list[KeyT] = []
        self.explicit_transaction = False
        self.cache = None
        self.scripts: set[Script[AnyStr]] = set()
        self.timeout = timeout
        self.type_adapter = client.type_adapter
        self._results: tuple[Any] | None = None

    @asynccontextmanager
    async def __asynccontextmanager__(self) -> AsyncGenerator[Self]:
        yield self
        await self._execute()

    def create_request(
        self,
        name: bytes,
        *arguments: ValueT,
        callback: Callable[..., T_co],
        execution_parameters: ExecutionParameters | None = None,
    ) -> CommandRequest[T_co]:
        """
        :meta private:
        """
        return ClusterPipelineCommandRequest(
            self, name, *arguments, callback=callback, execution_parameters=execution_parameters
        )

    @asynccontextmanager
    async def watch(self, *keys: KeyT) -> AsyncGenerator[None]:
        """
        The given keys will be watched for changes within this context and the
        commands stacked within the context will be automatically executed when the
        context exits.
        """
        if self.command_stack:
            raise WatchError("Unable to add a watch after pipeline commands have been added")
        self._watched_node = self.connection_pool.cluster_layout.node_for_request(b"WATCH", keys)
        self.watches.extend(keys)
        async with self.connection_pool.acquire(
            node=self._watched_node
        ) as self._watched_connection:
            await self._watched_connection.create_request(CommandName.WATCH, *keys)
            self.explicit_transaction = True
            yield
            await self._execute()
            await self._watched_connection.create_request(CommandName.UNWATCH, decode=False)

    @property
    def results(self) -> tuple[Any, ...] | None:
        """
        The results of the pipeline execution which can be accessed
        after the pipeline has completed.
        """
        if self.command_stack:
            raise RuntimeError("Pipeline results are not available before it completes")
        return self._results

    def execute_command(
        self,
        command: RedisCommandP,
        callback: Callable[..., R] = NoopCallback(),
        **options: Unpack[ExecutionParameters],
    ) -> Awaitable[R]:
        raise NotImplementedError

    async def _clear(self) -> None:
        """
        Clear the pipeline and reset state.
        """
        self.command_stack = []
        self.scripts.clear()
        self.watches.clear()
        self.explicit_transaction = False

    def _pipeline_execute_command(
        self,
        command: ClusterPipelineCommandRequest[Any],
    ) -> None:
        command.position = len(self.command_stack)
        self.command_stack.append(command)

    def _raise_first_error(self) -> None:
        for c in self.command_stack:
            r = c.result

            if isinstance(r, (RedisError, TimeoutError)):
                self._annotate_exception(r, c.position + 1, c.name, c.arguments)
                raise r

    def _annotate_exception(
        self,
        exception: RedisError | TimeoutError | None,
        number: int,
        command: bytes,
        args: Iterable[RedisValueT],
    ) -> None:
        if exception:
            cmd = command.decode("latin-1")
            args = " ".join(str(x) for x in args)
            msg = f"Command # {number} ({cmd} {args}) of pipeline caused error: {exception.args[0]}"
            exception.args = (msg,) + exception.args[1:]

    async def _execute(self) -> None:
        """
        Execute all queued commands in the cluster pipeline. Returns a tuple of results.
        """
        if not self.command_stack:
            return
        if self.scripts:
            await self._load_scripts()
        if self._transaction or self.explicit_transaction:
            execute = self._send_cluster_transaction
        else:
            execute = self._send_cluster_commands
        try:
            await execute(self._raise_on_error)
        finally:
            await self._clear()

    @retryable(policy=ConstantRetryPolicy((ClusterDownError,), retries=3, delay=0.1))
    async def _send_cluster_transaction(self, raise_on_error: bool = True) -> None:
        """
        :meta private:
        """
        attempt = sorted(self.command_stack, key=lambda x: x.position)
        slots: set[int] = set()
        for c in attempt:
            slots.add(self._determine_slot(c.name, *c.arguments, **c.execution_parameters))
            if len(slots) > 1:
                raise ClusterTransactionError("Multiple slots involved in transaction")
        if not slots:
            raise ClusterTransactionError("No slots found for transaction")
        node = self.connection_pool.cluster_layout.node_for_slot(slots.pop())
        if self._watched_node and node != self._watched_node:
            raise ClusterTransactionError("Multiple slots involved in transaction")

        node_commands = NodeCommands(
            self.client,
            node,
            in_transaction=True,
            timeout=self.timeout,
            connection=self._watched_connection,
            raise_on_error=self._raise_on_error,
        )
        node_commands.extend(attempt)
        async with node_commands:
            node_commands.write()
            try:
                await node_commands.read()
            except ExecAbortError:
                await node_commands.connection.create_request(CommandName.DISCARD)

            # If at least one watched key is modified before EXEC, the transaction aborts and EXEC returns null.
            if node_commands.exec_cmd:
                exec_result = await node_commands.exec_cmd
                if exec_result is None:
                    raise WatchError("Watched variable changed.")

            self._results = tuple(
                n.result
                for n in node_commands.commands
                if n.name not in {CommandName.MULTI, CommandName.EXEC}
            )
            if raise_on_error:
                self._raise_first_error()

    @retryable(policy=ConstantRetryPolicy((ClusterDownError,), retries=3, delay=0.1))
    async def _send_cluster_commands(
        self, raise_on_error: bool = True, allow_redirections: bool = True
    ) -> None:
        """
        Execute all queued commands in the cluster pipeline, handling redirections
        and retries as needed.

        :meta private:
        """
        # On first send, queue all commands. On retry, only failed ones.
        attempt: dict[ClusterPipelineCommandRequest[Any], ClusterNodeLocation] = {}
        # Group commands by node for efficient network usage.
        nodes: dict[str, NodeCommands] = {}
        for c in sorted(self.command_stack, key=lambda x: x.position):
            node = self.connection_pool.cluster_layout.node_for_request(c.name, c.arguments)

            if node.name not in nodes:
                nodes[node.name] = NodeCommands(
                    self.client,
                    node,
                    timeout=self.timeout,
                    raise_on_error=self._raise_on_error,
                )
            nodes[node.name].append(c)
            attempt[c] = node

        # Write to all nodes, then read from all nodes in sequence.
        for n in nodes.values():
            async with n:
                n.write()
                await n.read()

        # Retry MOVED/ASK/connection errors one by one if allowed.
        attempt = dict(
            sorted(
                (c for c in attempt.items() if isinstance(c[0].result, ERRORS_ALLOW_RETRY)),
                key=lambda x: x[0].position,
            )
        )

        if attempt and allow_redirections:
            for c, node in attempt.items():
                self.connection_pool.cluster_layout.report_errors(node, c.result)
                try:
                    c.result = await self.client.execute_command(
                        RedisCommand(c.name, c.arguments), **c.execution_parameters
                    )
                except (RedisError, TimeoutError) as e:
                    c.result = e

        # Flatten results to match the original command order.
        response = []
        for c in sorted(self.command_stack, key=lambda x: x.position):
            r = c.result
            if not isinstance(c.result, (RedisError, TimeoutError)):
                r = c.callback(c.result)
            c._response = PipelineResult(r)
            response.append(r)
        self._results = tuple(response)
        if raise_on_error:
            self._raise_first_error()

    def _determine_slot(
        self, command: bytes, *args: RedisValueT, **options: Unpack[ExecutionParameters]
    ) -> int:
        """
        Determine the hash slot for the given command and arguments.
        """
        keys: tuple[RedisValueT, ...] = (
            cast(tuple[RedisValueT, ...], options.get("keys"))
            or KeySpec.extract_keys(command, *args)[0]
        )

        if not keys:
            raise RedisClusterError(
                f"No way to dispatch {nativestr(command)} to Redis Cluster. Missing key"
            )
        slots = KeySpec.affected_slots(command, *args)

        if len(slots) != 1:
            raise ClusterCrossSlotError(command=command, keys=keys)
        return slots.pop()

    async def _load_scripts(self) -> None:
        shas = [s.sha for s in self.scripts]
        exists = await self.client.execute_command(
            RedisCommand(CommandName.SCRIPT_EXISTS, tuple(shas)), callback=BoolsCallback()
        )

        if not all(exists):
            for s, exist in zip(self.scripts, exists):
                if not exist:
                    s.sha = await self.client.execute_command(
                        RedisCommand(CommandName.SCRIPT_LOAD, (s.script,)),
                        callback=AnyStrCallback[AnyStr](),
                    )



from __future__ import annotations

from typing import cast

from coredis.response._callbacks import ResponseCallback
from coredis.response.types import ScoredMember, ScoredMembers
from coredis.typing import (
    AnyStr,
    Generic,
    ResponsePrimitive,
    ResponseType,
    StringT,
)


class ZRankCallback(
    ResponseCallback[
        int | list[ResponsePrimitive] | None,
        int | tuple[int, float] | None,
    ],
):
    def transform(
        self,
        response: int | list[ResponsePrimitive] | None,
    ) -> int | tuple[int, float] | None:
        if response is not None:
            if self.options.get("withscore"):
                return cast(tuple[int, float], tuple(cast(list[ResponsePrimitive], response)[:2]))
            else:
                return cast(int, response)
        return response


class ZMembersOrScoredMembers(
    ResponseCallback[
        list[AnyStr | list[ResponsePrimitive]],
        tuple[AnyStr | ScoredMember, ...],
    ],
):
    def transform(
        self,
        response: list[AnyStr | list[ResponsePrimitive]],
    ) -> tuple[AnyStr | ScoredMember, ...]:
        if self.options.get("withscores"):
            return tuple(ScoredMember(*v) for v in cast(list[tuple[AnyStr, float]], response))
        else:
            return cast(tuple[AnyStr, ...], tuple(response))


class ZSetScorePairCallback(
    ResponseCallback[
        list[ResponsePrimitive | list[ResponsePrimitive]] | None,
        ScoredMember | ScoredMembers | None,
    ],
    Generic[AnyStr],
):
    def transform(
        self,
        response: list[ResponsePrimitive | list[ResponsePrimitive]] | None,
    ) -> ScoredMember | ScoredMembers | None:
        if not response:
            return None

        if not (self.options.get("withscores") or self.options.get("count")):
            return ScoredMember(*cast(tuple[AnyStr, float], response))

        return tuple(ScoredMember(*v) for v in cast(list[tuple[AnyStr, float]], response))


class ZMPopCallback(
    ResponseCallback[
        list[ResponseType] | None,
        tuple[AnyStr, ScoredMembers] | None,
    ],
    Generic[AnyStr],
):
    def transform(
        self,
        response: list[ResponseType] | None,
    ) -> tuple[AnyStr, ScoredMembers] | None:
        r = cast(tuple[AnyStr, list[tuple[AnyStr, int]]], response)
        if r:
            return r[0], tuple(ScoredMember(v[0], float(v[1])) for v in r[1])

        return None


class ZScanCallback(
    ResponseCallback[list[ResponseType], tuple[int, ScoredMembers]],
    Generic[AnyStr],
):
    def transform(
        self,
        response: list[ResponseType],
    ) -> tuple[int, ScoredMembers]:
        cursor, r = cast(tuple[int, list[AnyStr]], response)
        it = iter(r)
        return int(cursor), tuple(ScoredMember(*v) for v in zip(it, map(float, it)))


class ZRandMemberCallback(
    ResponseCallback[
        AnyStr | list[list[StringT | float]] | list[StringT] | None,
        AnyStr | tuple[AnyStr, ...] | ScoredMembers | None,
    ]
):
    def transform(
        self,
        response: AnyStr | list[list[StringT | float]] | list[StringT] | None,
    ) -> AnyStr | tuple[AnyStr, ...] | ScoredMembers | None:
        if response:
            if not isinstance(response, list):
                return response

            if self.options.get("withscores"):
                return tuple(ScoredMember(*cast(tuple[StringT, float], tuple(v))) for v in response)
            else:
                return tuple(cast(list[AnyStr], response))
        return None


class BZPopCallback(
    ResponseCallback[
        list[StringT | float] | None,
        tuple[AnyStr, AnyStr, float] | None,
    ]
):
    def transform(
        self,
        response: list[StringT | float] | None,
    ) -> tuple[AnyStr, AnyStr, float] | None:
        if response:
            return cast(AnyStr, response[0]), cast(AnyStr, response[1]), float(response[2])
        return None


class ZAddCallback(ResponseCallback[int | float, int | float]):
    def transform(
        self,
        response: int | float,
    ) -> int | float:
        return response



from __future__ import annotations

from typing import cast

from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict
from coredis.typing import (
    AnyStr,
    ResponsePrimitive,
    StringT,
)


class HScanCallback(
    ResponseCallback[
        list[ResponsePrimitive | list[ResponsePrimitive]],
        tuple[int, dict[AnyStr, AnyStr] | tuple[AnyStr, ...]],
    ]
):
    def transform(
        self,
        response: list[ResponsePrimitive | list[ResponsePrimitive]],
    ) -> tuple[int, dict[AnyStr, AnyStr] | tuple[AnyStr, ...]]:
        cursor = int(cast(StringT, response[0]))
        results = cast(list[AnyStr], response[1])
        if self.options.get("novalues"):
            return cursor, tuple(results)
        else:
            return cursor, flat_pairs_to_dict(results)


class HRandFieldCallback(
    ResponseCallback[
        AnyStr | list[AnyStr] | list[list[AnyStr]] | None,
        AnyStr | tuple[AnyStr, ...] | dict[AnyStr, AnyStr] | None,
    ]
):
    def transform(
        self,
        response: AnyStr | list[AnyStr] | list[list[AnyStr]] | None,
    ) -> AnyStr | tuple[AnyStr, ...] | dict[AnyStr, AnyStr] | None:
        if not response:
            return None
        if self.options.get("count"):
            if self.options.get("withvalues"):
                return dict(cast(list[tuple[AnyStr, AnyStr]], response))
            return tuple(cast(tuple[AnyStr, AnyStr], response))
        return cast(AnyStr, response)


class HGetAllCallback(ResponseCallback[dict[AnyStr, AnyStr], dict[AnyStr, AnyStr]]):
    def transform(
        self,
        response: dict[AnyStr, AnyStr],
    ) -> dict[AnyStr, AnyStr]:
        return response



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import Any, cast, overload

from anyio import (
    TASK_STATUS_IGNORED,
    CapacityLimiter,
    fail_after,
    move_on_after,
)
from anyio.abc import TaskStatus

from coredis._concurrency import Queue
from coredis.connection._base import (
    BaseConnectionParams,
    ConnectionT,
    Location,
)
from coredis.connection._tcp import TCPConnection, TCPLocation
from coredis.connection._uds import UnixDomainSocketConnection, UnixDomainSocketLocation
from coredis.exceptions import ConnectionError, RedisError
from coredis.patterns.cache import AbstractCache, NodeTrackingCache, TrackingCache
from coredis.typing import (
    AsyncGenerator,
    Callable,
    ClassVar,
    NotRequired,
    Self,
    Unpack,
)

from ._base import BaseConnectionPool, BaseConnectionPoolParams


class ConnectionPoolParams(BaseConnectionPoolParams[ConnectionT]):
    """
    Parameters accepted by :class:`coredis.pool.ConnectionPool`
    """

    #: :meta private:
    _cache: NotRequired[AbstractCache | None]


class ConnectionPool(BaseConnectionPool[ConnectionT]):
    URL_QUERY_ARGUMENT_PARSERS: ClassVar[
        dict[str, Callable[..., int | float | bool | str | None]]
    ] = {
        **BaseConnectionPool.URL_QUERY_ARGUMENT_PARSERS,
    }

    @overload
    def __init__(
        self: ConnectionPool[ConnectionT],
        *,
        connection_class: type[ConnectionT],
        location: Location | None = ...,
        max_connections: int | None = ...,
        timeout: float | None = ...,
        _cache: AbstractCache | None = ...,
        **connection_kwargs: Unpack[BaseConnectionParams],
    ): ...
    @overload
    def __init__(
        self: ConnectionPool[TCPConnection],
        *,
        connection_class: None = ...,
        location: TCPLocation,
        max_connections: int | None = ...,
        timeout: float | None = ...,
        _cache: AbstractCache | None = ...,
        **connection_kwargs: Unpack[BaseConnectionParams],
    ): ...
    @overload
    def __init__(
        self: ConnectionPool[UnixDomainSocketConnection],
        *,
        connection_class: None = ...,
        location: UnixDomainSocketLocation,
        max_connections: int | None = ...,
        timeout: float | None = ...,
        _cache: AbstractCache | None = ...,
        **connection_kwargs: Unpack[BaseConnectionParams],
    ): ...
    @overload
    def __init__(
        self: ConnectionPool[TCPConnection],
        *,
        connection_class: None = ...,
        location: None = ...,
        # Retained for backward compatibility
        host: str = ...,
        port: int = ...,
        max_connections: int | None = ...,
        timeout: float | None = ...,
        _cache: AbstractCache | None = ...,
        **connection_kwargs: Unpack[BaseConnectionParams],
    ): ...

    def __init__(
        self,
        *,
        connection_class: type[ConnectionT] | None = None,
        location: Location | None = None,
        max_connections: int | None = None,
        timeout: float | None = None,
        _cache: AbstractCache | None = None,
        # host/port retained for backward compatibility
        host: str | None = None,
        port: int | None = None,
        **connection_kwargs: Unpack[BaseConnectionParams],
    ) -> None:
        """
        Blocking connection pool for single instance redis clients

        :param connection_class: The connection class to use when creating new connections
        :param max_connections: Maximum connections to grow the pool.
         Once the limit is reached clients will block to wait for a connection
         to be returned to the pool.
        :param timeout: Number of seconds to block when trying to obtain a connection.
        :param connection_kwargs: arguments to pass to the :paramref:`connection_class`
         constructor when creating a new connection
        """
        if connection_class is None:
            if isinstance(location, TCPLocation):
                connection_class = cast(type[ConnectionT], TCPConnection)
            elif isinstance(location, UnixDomainSocketLocation):
                connection_class = cast(type[ConnectionT], UnixDomainSocketConnection)
            elif host is not None and port is not None:
                connection_class = cast(type[ConnectionT], TCPConnection)
                location = TCPLocation(host, port)
        if not connection_class:
            raise RuntimeError("Unable to initialize pool without a `connection_class`")
        super().__init__(
            connection_class=connection_class,
            location=location,
            max_connections=max_connections,
            timeout=timeout,
            **connection_kwargs,
        )
        # TODO: Use the `max_failures` argument of tracking cache
        self.cache: TrackingCache[Any] | None = NodeTrackingCache(self, _cache) if _cache else None
        # The pool of available connections
        self._available_connections: Queue[ConnectionT] = Queue(self.max_connections)
        # All connections that are still active
        self._online_connections: set[ConnectionT] = set()
        # This should be used by the connection to limit concurrently entering
        # CPU hotspots to ensure all fairness between connections in the pool.
        # The main observed scenario where this can happen is if the connection pool
        # is being used by multiple push message consumers that are constantly
        # receiving data in the read task.
        self._connection_processing_budget = CapacityLimiter(1)
        self.connection_kwargs["processing_budget"] = self._connection_processing_budget

    @overload
    @classmethod
    def from_url(
        cls: type[ConnectionPool[Any]],
        url: str,
        *,
        decode_components: bool = False,
        **kwargs: Unpack[ConnectionPoolParams[Any]],
    ) -> ConnectionPool[TCPConnection] | ConnectionPool[UnixDomainSocketConnection]: ...

    @overload
    @classmethod
    def from_url(
        cls: type[Self],
        url: str,
        *,
        decode_components: bool = False,
        **kwargs: Unpack[ConnectionPoolParams[Any]],
    ) -> Self: ...

    @classmethod
    def from_url(
        cls: type[Self],
        url: str,
        *,
        decode_components: bool = False,
        **kwargs: Unpack[ConnectionPoolParams[Any]],
    ) -> Self | ConnectionPool[TCPConnection] | ConnectionPool[UnixDomainSocketConnection]:
        """
        Returns a connection pool configured from the given URL.

        For example:

        - ``redis://[:password]@localhost:6379/0``
        - ``rediss://[:password]@localhost:6379/0``
        - ``unix://[:password]@/path/to/socket.sock?db=0``

        Three URL schemes are supported:

        - `redis:// <http://www.iana.org/assignments/uri-schemes/prov/redis>`__
          creates a normal TCP socket connection
        - `rediss:// <http://www.iana.org/assignments/uri-schemes/prov/rediss>`__
          creates a SSL wrapped TCP socket connection
        - ``unix://`` creates a Unix Domain Socket connection

        There are several ways to specify a database number. The parse function
        will return the first specified option:

        - A ``db`` querystring option, e.g. ``redis://localhost?db=0``
        - If using the ``redis://`` scheme, the path argument of the url, e.g.
          ``redis://localhost/0``
        - The ``db`` argument to this function.

        If none of these options are specified, ``db=0`` is used.

        The :paramref:`decode_components` argument allows this function to work with
        percent-encoded URLs. If this argument is set to ``True`` all ``%xx``
        escapes will be replaced by their single-character equivalents after
        the URL has been parsed. This only applies to the ``hostname``,
        ``path``, ``username`` & ``password`` components.

        Any additional querystring arguments and keyword arguments will be
        passed along to the class constructor.

        .. note:: In the case of conflicting arguments, querystring arguments always win.
        """
        location, merged_options = cls._parse_url(
            url, decode_components, kwargs, ConnectionPoolParams
        )
        if isinstance(location, UnixDomainSocketLocation):
            merged_options["connection_class"] = UnixDomainSocketConnection
        else:
            merged_options["connection_class"] = TCPConnection
        return cls(
            location=location,
            **merged_options,
        )

    async def _initialize(self) -> None:
        if self.cache:
            with move_on_after(self.connect_timeout) as cancel_scope:
                await self.task_group.start(self.cache.run)
            if cancel_scope.cancelled_caught:
                raise ConnectionError(
                    f"Unable to initialize cache within {self.connect_timeout} seconds"
                )

    async def get_connection(self, **_: Any) -> ConnectionT:
        """
        Gets or create a connection from the pool. Be careful to only release the
        connection AFTER all commands are sent, or race conditions are possible.
        """
        with fail_after(self.timeout):
            # if stack has a connection, use that
            connection = await self._available_connections.get()
            if connection is None or not connection.reusable:
                # If the connection was in the pool but is "dirty" it should be
                # invalidated (to avoid leaking) and discarded.
                if connection:
                    connection.invalidate()
                connection = await self._construct_connection()
                if err := await self.task_group.start(self.__wrap_connection, connection):
                    self._available_connections.append_nowait(None)
                    raise err
                self._online_connections.add(connection)
            return connection

    @asynccontextmanager
    async def acquire(self, **_: Any) -> AsyncGenerator[ConnectionT]:
        """
        Gets or creates a connection from the pool, then release it afterwards.
        Multiplexing is automatic if you exit the context manager before
        waiting for command results.

        .. caution:: Do not explicitly release connections acquired
           using this context manager.
        """
        connection = await self.get_connection()
        yield connection
        self.release(connection)

    def release(self, connection: ConnectionT) -> None:
        """
        Checks connection for liveness and releases it back to the pool.
        """
        if connection.usable:
            self._available_connections.put_nowait(connection)

    def disconnect(self) -> None:
        """
        Disconnect all active connections in the pool
        """
        for connection in self._online_connections:
            connection.invalidate()
        self._online_connections.clear()

    def _reset(self) -> None:
        # TODO: seems like something should be cleared?
        pass

    async def _construct_connection(self) -> ConnectionT:
        assert self.location
        return self.connection_class(self.location, **self.connection_kwargs)

    async def __wrap_connection(
        self,
        connection: ConnectionT,
        *,
        task_status: TaskStatus[None | Exception] = TASK_STATUS_IGNORED,
    ) -> None:
        try:
            await connection.run(task_status=task_status)
        except RedisError as error:
            # Only coredis.exception.RedisError is explictly caught and returned with the task status
            # As these are clear signals that an error case was handled by the connection
            task_status.started(error)
        finally:
            self._online_connections.discard(connection)
            if connection in self._available_connections:
                self._available_connections.remove(connection)
                self._available_connections.append_nowait(None)



from __future__ import annotations

import warnings
from contextlib import AsyncExitStack, asynccontextmanager
from typing import Any, overload

from anyio import AsyncContextManagerMixin

from coredis._utils import nativestr
from coredis.connection._tcp import TCPLocation
from coredis.exceptions import (
    ConnectionError,
    PrimaryNotFoundError,
    ResponseError,
)
from coredis.patterns.cache import AbstractCache
from coredis.pool._sentinel import SentinelConnectionPool
from coredis.retry import NoRetryPolicy, RetryPolicy
from coredis.typing import (
    AnyStr,
    AsyncGenerator,
    Generic,
    Iterable,
    Literal,
    ResponsePrimitive,
    Self,
    TypeAdapter,
)

from .basic import Redis


class Sentinel(AsyncContextManagerMixin, Generic[AnyStr]):
    """
    Example use::

        from coredis import Sentinel
        from coredis.connection import TCPLocation
        async def test():
            async with Sentinel([TCPLocation('localhost', 26379)], stream_timeout=0.1) as sentinel:
                primary = sentinel.primary_for('my-instance', stream_timeout=0.1)
                replica = sentinel.replica_for('my-instance', stream_timeout=0.1)
                async with primary, replica:
                    await primary.set('foo', 'bar')
                    await replica.get('foo')

    """

    @overload
    def __init__(
        self: Sentinel[bytes],
        sentinels: Iterable[TCPLocation | tuple[str, int]],
        min_other_sentinels: int = ...,
        sentinel_kwargs: dict[str, Any] | None = ...,
        decode_responses: Literal[False] = ...,
        cache: AbstractCache | None = None,
        type_adapter: TypeAdapter | None = ...,
        retry_policy: RetryPolicy = ...,
        **connection_kwargs: Any,
    ) -> None: ...

    @overload
    def __init__(
        self: Sentinel[str],
        sentinels: Iterable[TCPLocation | tuple[str, int]],
        min_other_sentinels: int = ...,
        sentinel_kwargs: dict[str, Any] | None = ...,
        decode_responses: Literal[True] = ...,
        cache: AbstractCache | None = None,
        type_adapter: TypeAdapter | None = None,
        retry_policy: RetryPolicy = ...,
        **connection_kwargs: Any,
    ) -> None: ...

    def __init__(
        self,
        sentinels: Iterable[TCPLocation | tuple[str, int]],
        min_other_sentinels: int = 0,
        sentinel_kwargs: dict[str, Any] | None = None,
        decode_responses: bool = False,
        cache: AbstractCache | None = None,
        type_adapter: TypeAdapter | None = None,
        retry_policy: RetryPolicy = NoRetryPolicy(),
        **connection_kwargs: Any,
    ) -> None:
        """
        Changes
          - .. versionadded:: 3.10.0
               Accept :paramref:`cache` parameter to be used with primaries
               and replicas returned from the sentinel instance.

        :param sentinels: is a list of sentinel nodes. Each node is represented by
         a pair (hostname, port).
        :param min_other_sentinels: defined a minimum number of peers for a sentinel.
         When querying a sentinel, if it doesn't meet this threshold, responses
         from that sentinel won't be considered valid.
        :param sentinel_kwargs: is a dictionary of connection arguments used when
         connecting to sentinel instances. Any argument that can be passed to
         a normal Redis connection can be specified here. If :paramref:`sentinel_kwargs` is
         not specified, ``stream_timeout``, ``socket_keepalive`` and ``decode_responses``
         options specified in :paramref:`connection_kwargs` will be used.
        :param cache: If provided the cache will be shared between both primaries and replicas
         returned by this sentinel.
        :param type_adapter: The adapter to use for serializing / deserializing customs types
         when interacting with redis commands. If provided this adapter will be used for both
         primaries and replicas returned by this sentinel.
        :param retry_policy: The retry policy to use when interacting with the the primary
         and replica instances.
        :param connection_kwargs: are keyword arguments that will be used when
         establishing a connection to a Redis server (i.e. are passed on to the
         constructor of :class:`Redis` for all primary and replicas).
        """
        # if sentinel_kwargs isn't defined, use the socket_* options from
        # connection_kwargs
        if sentinel_kwargs is None:
            sentinel_kwargs = {
                k: v
                for k, v in connection_kwargs.items()
                if k
                in {
                    "connect_timeout",
                    "socket_timeout",
                    "socket_keepalive",
                    "encoding",
                }
            }
        self.sentinel_kwargs = sentinel_kwargs
        self.min_other_sentinels = min_other_sentinels
        self.connection_kwargs = connection_kwargs
        self.__cache = cache
        self.__type_adapter = type_adapter
        self.__retry_policy = retry_policy
        self.connection_kwargs["decode_responses"] = self.sentinel_kwargs["decode_responses"] = (
            decode_responses
        )
        if any(isinstance(sentinel, tuple) for sentinel in sentinels):
            warnings.warn(
                "Use coredis.connection.TCPLocation to specify sentinels",
                DeprecationWarning,
                stacklevel=2,
            )
        sentinel_locations = [
            location if isinstance(location, TCPLocation) else TCPLocation(*location)
            for location in sentinels
        ]
        self.sentinels = [
            Redis(
                location.host,
                location.port,
                retry_policy=self.__retry_policy,
                **self.sentinel_kwargs,
            )
            for location in sentinel_locations
        ]

    @asynccontextmanager
    async def __asynccontextmanager__(self) -> AsyncGenerator[Self]:
        async with AsyncExitStack() as stack:
            for sentinel in self.sentinels:
                await stack.enter_async_context(sentinel)
            yield self

    def __repr__(self) -> str:
        sentinels = [
            f"{location}"
            for sentinel in self.sentinels
            if (location := sentinel.connection_pool.location)
        ]

        return f"Sentinel<sentinels=[{','.join(sentinels)}]>"

    def __filter_replicas(
        self, replicas: Iterable[dict[str, ResponsePrimitive]]
    ) -> list[tuple[str, int]]:
        """Removes replicas that are in an ODOWN or SDOWN state"""
        replicas_alive: list[tuple[str, int]] = []
        for replica in replicas:
            if replica["is_odown"] or replica["is_sdown"] or replica["is_disconnected"]:
                continue
            ip, port = replica["ip"], replica["port"]
            assert ip is not None and port is not None
            replicas_alive.append((nativestr(ip), int(port)))
        return replicas_alive

    async def discover_primary(self, service_name: str) -> tuple[str, int]:
        """
        Asks sentinel servers for the Redis primary's location corresponding
        to the service labeled :paramref:`service_name`.

        :return: A pair (location, port) or raises :exc:`~coredis.exceptions.PrimaryNotFoundError`
         if no primary is found.
        """

        for idx, sentinel in enumerate(self.sentinels):
            try:
                primary_location = await sentinel.sentinel_get_master_addr_by_name(service_name)
            except (ConnectionError, TimeoutError):
                continue
            if not primary_location:
                continue
            # Put this sentinel at the top of the list
            sentinel = self.sentinels.pop(idx)
            self.sentinels.insert(0, sentinel)
            return nativestr(primary_location[0]), primary_location[1]
        raise PrimaryNotFoundError(f"No primary found for {service_name!r}")

    async def discover_replicas(self, service_name: str) -> list[tuple[str, int]]:
        """Returns a list of alive replicas for service :paramref:`service_name`"""
        for sentinel in self.sentinels:
            try:
                replicas = await sentinel.sentinel_replicas(service_name)
            except (ConnectionError, ResponseError, TimeoutError):
                continue
            return self.__filter_replicas(replicas)
        return []

    @overload
    def primary_for(
        self: Sentinel[bytes],
        service_name: str,
        *,
        redis_class: type[Redis[bytes]] = ...,
        connection_pool_class: type[SentinelConnectionPool] = ...,
        **kwargs: Any,
    ) -> Redis[bytes]: ...

    @overload
    def primary_for(
        self: Sentinel[str],
        service_name: str,
        *,
        redis_class: type[Redis[str]] = ...,
        connection_pool_class: type[SentinelConnectionPool] = ...,
        **kwargs: Any,
    ) -> Redis[str]: ...

    def primary_for(
        self,
        service_name: str,
        *,
        redis_class: type[Redis[Any]] = Redis[Any],
        connection_pool_class: type[SentinelConnectionPool] = SentinelConnectionPool,
        **kwargs: Any,
    ) -> Redis[bytes] | Redis[str]:
        """
        Returns a redis client instance for the :paramref:`service_name` primary.

        A :class:`coredis.sentinel.SentinelConnectionPool` class is used to
        retrive the primary's location before establishing a new connection.

        NOTE: If the primary's location has changed, any cached connections to
        the old primary are closed.

        By default clients will be a :class:`~coredis.Redis` instances.
        Specify a different class to the :paramref:`redis_class` argument if you desire
        something different.

        The :paramref:`connection_pool_class` specifies the connection pool to use.
        The :class:`~coredis.sentinel.SentinelConnectionPool` will be used by default.

        All other keyword arguments are merged with any :paramref:`Sentinel.connection_kwargs`
        passed to this class and passed to the connection pool as keyword
        arguments to be used to initialize Redis connections.
        """
        kwargs["is_primary"] = True
        connection_kwargs = dict(self.connection_kwargs)
        connection_kwargs.update(kwargs)

        return redis_class(
            connection_pool=connection_pool_class(
                service_name,
                self,
                _cache=self.__cache,
                **connection_kwargs,
            ),
            retry_policy=self.__retry_policy,
            type_adapter=self.__type_adapter,
        )

    @overload
    def replica_for(
        self: Sentinel[bytes],
        service_name: str,
        redis_class: type[Redis[bytes]] = ...,
        connection_pool_class: type[SentinelConnectionPool] = ...,
        **kwargs: Any,
    ) -> Redis[bytes]: ...

    @overload
    def replica_for(
        self: Sentinel[str],
        service_name: str,
        redis_class: type[Redis[str]] = ...,
        connection_pool_class: type[SentinelConnectionPool] = ...,
        **kwargs: Any,
    ) -> Redis[str]: ...

    def replica_for(
        self,
        service_name: str,
        redis_class: type[Redis[Any]] = Redis[Any],
        connection_pool_class: type[SentinelConnectionPool] = SentinelConnectionPool,
        **kwargs: Any,
    ) -> Redis[bytes] | Redis[str]:
        """
        Returns redis client instance for the :paramref:`service_name` replica(s).

        A SentinelConnectionPool class is used to retrieve the replica's
        location before establishing a new connection.

        By default clients will be a redis.Redis instance. Specify a
        different class to the :paramref:`redis_class` argument if you desire
        something different.

        The :paramref:`connection_pool_class` specifies the connection pool to use.
        The SentinelConnectionPool will be used by default.

        All other keyword arguments are merged with any :paramref:`Sentinel.connection_kwargs`
        passed to this class and passed to the connection pool as keyword
        arguments to be used to initialize Redis connections.
        """
        kwargs["is_primary"] = False
        connection_kwargs = dict(self.connection_kwargs)
        connection_kwargs.update(kwargs)

        return redis_class(
            connection_pool=connection_pool_class(
                service_name,
                self,
                _cache=self.__cache,
                **connection_kwargs,
            ),
            retry_policy=self.__retry_policy,
            type_adapter=self.__type_adapter,
        )



from __future__ import annotations

from typing import cast

from coredis._json import json
from coredis.response._callbacks import ResponseCallback
from coredis.typing import JsonType, ResponseType


class JsonCallback(ResponseCallback[ResponseType, JsonType]):
    def transform(
        self,
        response: ResponseType,
    ) -> JsonType:
        if isinstance(response, (bytes, str)):
            deser = json.loads(response)
        elif isinstance(response, list):
            deser = [json.loads(e) if isinstance(e, (bytes, str)) else e for e in response]
        else:
            deser = response
        return cast(JsonType, deser)  # alas we lie.



from __future__ import annotations

from coredis.response._callbacks import NoopCallback
from coredis.typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    ExecutionParameters,
    Protocol,
    R,
    RedisCommandP,
    TypeVar,
    Unpack,
    ValueT,
)

T_co = TypeVar("T_co", covariant=True)


if TYPE_CHECKING:
    from coredis.commands import CommandRequest


class AbstractExecutor(Protocol):
    def execute_command(
        self,
        command: RedisCommandP,
        callback: Callable[..., R] = NoopCallback(),
        **options: Unpack[ExecutionParameters],
    ) -> Awaitable[R]: ...

    def create_request(
        self,
        name: bytes,
        *arguments: ValueT,
        callback: Callable[..., R],
        execution_parameters: ExecutionParameters | None = None,
    ) -> CommandRequest[R]: ...



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any

from coredis.commands._routing import RoutingStrategy
from coredis.commands.constants import CommandFlag

if TYPE_CHECKING:
    from coredis.modules.base import ModuleGroupRegistry, ModuleRegistry

#: Populated by the @redis_command wrapper
READONLY_COMMANDS: set[bytes] = set()
#: Populated by the @redis_command wrapper
COMMAND_FLAGS: dict[bytes, set[CommandFlag]] = defaultdict(set)

#: Populated by the @redis_command wrapper
ROUTING_STRATEGIES: dict[bytes, RoutingStrategy[Any]] = {}

#: Populated by the @redis_command wrapper
CACHEABLE_COMMANDS: set[bytes] = set()

#: Populated by ModuleGroupRegistry
MODULE_GROUPS: set[ModuleGroupRegistry] = set()

#: Populated by ModuleRegistry
MODULES: dict[str, ModuleRegistry] = {}



from __future__ import annotations

from datetime import datetime

from coredis.exceptions import DataError, NoKeyError, RedisError
from coredis.response._callbacks import DateTimeCallback, ResponseCallback
from coredis.typing import (
    AnyStr,
    ResponseType,
    StringT,
    TypeGuard,
)


class SortCallback(
    ResponseCallback[
        int | list[AnyStr],
        int | tuple[AnyStr, ...],
    ]
):
    def transform(
        self,
        response: int | list[AnyStr],
    ) -> int | tuple[AnyStr, ...]:
        if isinstance(response, list):
            return tuple(response)
        return response


class ScanCallback(ResponseCallback[list[ResponseType], tuple[int, tuple[AnyStr, ...]]]):
    def guard(self, response: list[ResponseType]) -> TypeGuard[tuple[StringT, list[AnyStr]]]:
        return isinstance(response[0], (str, bytes)) and isinstance(response[1], list)

    def transform(
        self,
        response: list[ResponseType],
    ) -> tuple[int, tuple[AnyStr, ...]]:
        assert self.guard(response)
        cursor, r = response
        return int(cursor), tuple(r)


class ExpiryCallback(DateTimeCallback):
    def transform(
        self,
        response: int | float,
    ) -> datetime:
        if response > 0:
            return super().transform(response)
        else:
            if response == -2:
                raise NoKeyError()
            elif response == -1:
                raise DataError()
            else:
                raise RedisError()



"""
coredis.response.callbacks
--------------------------
"""

from __future__ import annotations

import datetime
import itertools
from abc import ABC, abstractmethod
from typing import Any, cast

from coredis._utils import b
from coredis.exceptions import ClusterResponseError, ResponseError
from coredis.typing import (
    AnyStr,
    Callable,
    Generic,
    Iterable,
    ParamSpec,
    RedisValueT,
    ResponsePrimitive,
    ResponseType,
    Sequence,
    StringT,
    TypeGuard,
    TypeVar,
    add_runtime_checks,
)

R = TypeVar("R")
S = TypeVar("S")
T = TypeVar("T")
P = ParamSpec("P")
CR_co = TypeVar("CR_co", covariant=True)
CK_co = TypeVar("CK_co", covariant=True)
RESP3 = TypeVar("RESP3")


class ResponseCallback(ABC, Generic[RESP3, R]):
    def __init__(self, **options: Any) -> None:
        self.options = options

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        setattr(cls, "transform", add_runtime_checks(getattr(cls, "transform")))

    def __call__(
        self,
        response: RESP3,
    ) -> R:
        return self.transform(response)

    @abstractmethod
    def transform(self, response: RESP3) -> R:
        pass


class NoopCallback(ResponseCallback[R, R]):
    def transform(self, response: R) -> R:
        return response


class ClusterMultiNodeCallback(ABC, Generic[R]):
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        setattr(cls, "combine", add_runtime_checks(getattr(cls, "combine")))

    def __call__(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[R | ResponseError | TimeoutError],
    ) -> R:
        if self.successful(responses):
            return self.combine(key_positions, responses)
        else:
            return self.handle_error(key_positions, responses)

    @property
    @abstractmethod
    def response_policy(self) -> str: ...

    @abstractmethod
    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[R],
    ) -> R:
        pass

    def handle_error(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[R | ResponseError | TimeoutError],
    ) -> R:
        for response in responses:
            if isinstance(response, (ResponseError, TimeoutError)):
                raise response
        assert False

    def successful(self, values: list[R | ResponseError | TimeoutError]) -> TypeGuard[list[R]]:
        for value in values:
            if isinstance(value, (ResponseError, TimeoutError)):
                return False
        return True


class ClusterNoMerge(ClusterMultiNodeCallback[R]):
    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[R],
    ) -> R:
        assert len(responses) == 1
        return responses[0]

    @property
    def response_policy(self) -> str:
        return "the response from the node"


class ClusterBoolCombine(ClusterMultiNodeCallback[bool]):
    def __init__(self, any: bool = False):
        self.any = any

    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[bool],
    ) -> bool:
        assert (isinstance(response, bool) for response in responses)
        return any(responses) if self.any else all(responses)

    @property
    def response_policy(self) -> str:
        return (
            "``True`` if any response is``True``"
            if self.any
            else "``True`` if all responses are ``True``"
        )


class ClusterAlignedBoolsCombine(ClusterMultiNodeCallback[tuple[bool, ...]]):
    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[tuple[bool, ...]],
    ) -> tuple[bool, ...]:
        return tuple(all(k) for k in zip(*responses))

    @property
    def response_policy(self) -> str:
        return "the logical AND of all responses"


class ClusterEnsureConsistent(ClusterMultiNodeCallback[R | None]):
    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[R | None],
    ) -> R | None:
        if len(set(responses)) != 1:
            raise ClusterResponseError("Inconsistent response from cluster nodes")
        elif responses:
            return responses[0]
        return None

    @property
    def response_policy(self) -> str:
        return "the first response if all responses are consistent"


class ClusterFirstNonException(ClusterMultiNodeCallback[R | None]):
    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[R | None],
    ) -> R | None:
        if responses:
            return responses[0]
        return None

    def handle_error(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[R | None | ResponseError | TimeoutError],
    ) -> R | None:
        first_exc: ResponseError | TimeoutError | None = None
        for response in responses:
            if not isinstance(response, (ResponseError, TimeoutError)):
                return response
            else:
                first_exc = response
        if first_exc is not None:
            raise first_exc
        return None

    @property
    def response_policy(self) -> str:
        return "the first response that is not an error"


class ClusterMergeSets(ClusterMultiNodeCallback[set[R]]):
    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[set[R]],
    ) -> set[R]:
        return set(itertools.chain(*responses))

    @property
    def response_policy(self) -> str:
        return "the union of the results"


class ClusterSum(ClusterMultiNodeCallback[int]):
    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[int],
    ) -> int:
        return sum(responses)

    @property
    def response_policy(self) -> str:
        return "the sum of results"


class ClusterMergeMapping(ClusterMultiNodeCallback[dict[R, S]]):
    def __init__(self, value_combine: Callable[[Iterable[S]], S]) -> None:
        self.value_combine = value_combine

    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[dict[R, S]],
    ) -> dict[R, S]:
        response: dict[R, S] = {}
        for key in set(itertools.chain(*responses)):
            values = list(
                response[key] for idx, response in enumerate(responses) if key in response
            )
            response[key] = self.value_combine(values)
        return response

    @property
    def response_policy(self) -> str:
        return "the merged mapping"


class ClusterConcatenateTuples(ClusterMultiNodeCallback[tuple[R, ...]]):
    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[tuple[R, ...]],
    ) -> tuple[R, ...]:
        total = sum(len(response) for response in responses)
        output: list[R | None] = [None] * total
        for chunk_key_positions, result in zip(key_positions, responses):
            for position, value in zip(chunk_key_positions, result):
                output[position] = value
        return tuple(cast(list[R], output))

    @property
    def response_policy(self) -> str:
        return "the concatenations of the results"


class ClusterConcatenateLists(ClusterMultiNodeCallback[list[R]]):
    def combine(
        self,
        key_positions: list[tuple[int, ...]],
        responses: list[list[R]],
    ) -> list[R]:
        total = sum(len(response) for response in responses)
        output: list[R | None] = [None] * total
        for chunk_key_positions, result in zip(key_positions, responses):
            for position, value in zip(chunk_key_positions, result):
                output[position] = value
        return cast(list[R], output)

    @property
    def response_policy(self) -> str:
        return "the concatenations of the results"


class SimpleStringCallback(ResponseCallback[StringT | None, bool]):
    def __init__(
        self,
        raise_on_error: type[Exception] | None = None,
        prefix_match: bool = False,
        ok_values: set[str] = {"OK"},
        **options: Any,
    ):
        self.raise_on_error = raise_on_error
        self.prefix_match = prefix_match
        self.ok_values = {b(v) for v in ok_values}
        super().__init__(**options)

    def transform(self, response: StringT | None, **options: Any) -> bool:
        if response:
            if not self.prefix_match:
                success = b(response) in self.ok_values
            else:
                success = any(b(response).startswith(ok) for ok in self.ok_values)
        else:
            success = False
        if not success and self.raise_on_error:
            raise self.raise_on_error(response)
        return success


class IntCallback(ResponseCallback[int, int]):
    def transform(self, response: ResponsePrimitive, **options: Any) -> int:
        if isinstance(response, int):
            return response
        raise ValueError(f"Unable to map {response!r} to int")


class AnyStrCallback(ResponseCallback[StringT, AnyStr]):
    def transform(self, response: StringT, **options: Any) -> AnyStr:
        if isinstance(response, (bytes, str)):
            return cast(AnyStr, response)

        raise ValueError(f"Unable to map {response!r} to AnyStr")


class FloatCallback(ResponseCallback[StringT | int | float, float]):
    def transform(self, response: StringT | int | float, **options: Any) -> float:
        if isinstance(response, float):
            return response
        if isinstance(response, (int, bytes, str)):
            return float(response)

        raise ValueError(f"Unable to map {response} to float")


class BoolCallback(ResponseCallback[int | bool, bool]):
    def transform(self, response: int | bool, **options: Any) -> bool:
        if isinstance(response, bool):
            return response
        return bool(response)


class SimpleStringOrIntCallback(ResponseCallback[RedisValueT, bool | int]):
    def transform(self, response: RedisValueT, **options: Any) -> bool | int:
        if isinstance(response, (int, bool)):
            return response
        elif isinstance(response, (str, bytes)):
            return SimpleStringCallback().transform(response)
        raise ValueError(f"Unable to map {response!r} to bool")


class TupleCallback(ResponseCallback[list[ResponseType], tuple[R, ...]]):
    def transform(self, response: ResponseType, **options: Any) -> tuple[R, ...]:
        if isinstance(response, list):
            return cast(tuple[R, ...], tuple(response))
        raise ValueError(f"Unable to map {response!r} to tuple")


class ItemOrTupleCallback(
    ResponseCallback[
        list[ResponseType] | ResponsePrimitive,
        tuple[R, ...] | R,
    ]
):
    def transform(
        self, response: list[ResponseType] | ResponsePrimitive, **options: Any
    ) -> tuple[R, ...] | R:
        if isinstance(response, list):
            return cast(tuple[R, ...], tuple(response))
        return cast(R, response)


class MixedTupleCallback(ResponseCallback[list[ResponseType], tuple[R, S]]):
    def transform(self, response: ResponseType, **options: Any) -> tuple[R, S]:
        if isinstance(response, list):
            return cast(tuple[R, S], tuple(response))
        raise ValueError(f"Unable to map {response!r} to tuple")


class ListCallback(ResponseCallback[list[ResponseType], list[R]]):
    def transform(self, response: list[ResponseType], **options: Any) -> list[R]:
        return cast(list[R], response)


class DateTimeCallback(ResponseCallback[int | float, datetime.datetime]):
    def transform(
        self,
        response: int | float,
    ) -> datetime.datetime:
        ts = float(response) if not isinstance(response, float) else response
        if self.options.get("unit") == "milliseconds":
            ts = ts / 1000.0
        return datetime.datetime.fromtimestamp(ts)


class DictCallback(
    ResponseCallback[
        Sequence[ResponseType] | dict[ResponsePrimitive, ResponseType],
        dict[R, S],
    ]
):
    def transform(
        self,
        response: Sequence[ResponseType] | dict[ResponsePrimitive, ResponseType],
        **options: Any,
    ) -> dict[R, S]:
        if isinstance(response, dict):
            return cast(dict[R, S], response)
        elif isinstance(response, list):
            it = iter(response)
            return cast(dict[R, S], dict(zip(it, it)))
        raise ValueError(f"Unable to map {response!r} to mapping")


class SetCallback(
    ResponseCallback[
        list[ResponsePrimitive] | set[ResponsePrimitive],
        set[R],
    ]
):
    def transform(
        self,
        response: list[ResponsePrimitive] | set[ResponsePrimitive],
        **options: Any,
    ) -> set[R]:
        if isinstance(response, list):
            return cast(set[R], set(response))
        elif isinstance(response, set):
            return cast(set[R], response)
        raise ValueError(f"Unable to map {response} to set")


class OneOrManyCallback(
    ResponseCallback[
        CR_co | list[CR_co | None] | None,
        CR_co | list[CR_co | None] | None,
    ]
):
    def transform(
        self,
        response: CR_co | list[CR_co | None] | None,
        **options: Any,
    ) -> CR_co | list[CR_co | None] | None:
        return response


class BoolsCallback(ResponseCallback[Iterable[int | bool], tuple[bool, ...]]):
    def transform(self, response: Iterable[int | bool], **options: Any) -> tuple[bool, ...]:
        if isinstance(response, list):
            return tuple(BoolCallback().transform(r) for r in response)
        return ()


class FloatsCallback(ResponseCallback[list[StringT | float | int], tuple[float, ...]]):
    def transform(self, response: list[StringT | float | int], **options: Any) -> tuple[float, ...]:
        if isinstance(response, list):
            return tuple(FloatCallback().transform(r) for r in response)
        return ()


class OptionalFloatCallback(
    ResponseCallback[
        StringT | int | float | None,
        float | None,
    ]
):
    def transform(
        self,
        response: StringT | int | float | None,
        **options: Any,
    ) -> float | None:
        if response is None:
            return None
        return FloatCallback().transform(response)


class OptionalIntCallback(ResponseCallback[int | None, int | None]):
    def transform(self, response: int | None, **options: Any) -> int | None:
        if response is None:
            return None
        if isinstance(response, int):
            return response
        raise ValueError(f"Unable to map {response} to int")


class OptionalAnyStrCallback(
    ResponseCallback[
        StringT | None,
        AnyStr | None,
    ]
):
    def transform(self, response: StringT | None, **options: Any) -> AnyStr | None:
        if response is None:
            return None
        if isinstance(response, (bytes, str)):
            return cast(AnyStr, response)
        raise ValueError(f"Unable to map {response} to AnyStr")


class OptionalListCallback(ResponseCallback[list[ResponseType], list[CR_co] | None]):
    def transform(self, response: ResponseType, **options: Any) -> list[CR_co] | None:
        return cast(list[CR_co], response)


class FirstValueCallback(ResponseCallback[dict[R, S], S]):
    def transform(self, response: dict[R, S], **options: Any) -> S:
        if response:
            return list(response.values())[0]
        else:
            raise ValueError("Empty response")



"""
coredis
-------

coredis is an async redis client with support for redis server,
cluster & sentinel.
"""

from __future__ import annotations

from coredis._version import __version__
from coredis.client import Redis, RedisCluster, Sentinel
from coredis.config import Config
from coredis.connection import (
    BaseConnection,
    ClusterConnection,
    Connection,
    SentinelManagedConnection,
    TCPConnection,
    UnixDomainSocketConnection,
)
from coredis.pool import (
    BaseConnectionPool,
    ClusterConnectionPool,
    ConnectionPool,
    SentinelConnectionPool,
)
from coredis.tokens import PureToken

__all__ = [
    "Config",
    "Redis",
    "RedisCluster",
    "BaseConnection",
    "Connection",
    "TCPConnection",
    "UnixDomainSocketConnection",
    "BaseConnectionPool",
    "ClusterConnection",
    "ConnectionPool",
    "ClusterConnectionPool",
    "PureToken",
    "Sentinel",
    "SentinelManagedConnection",
    "SentinelConnectionPool",
    "__version__",
]



from __future__ import annotations

import dataclasses
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from coredis.exceptions import ResponseError
from coredis.response._callbacks import ClusterMultiNodeCallback, ClusterNoMerge
from coredis.typing import (
    ClassVar,
    ExecutionParameters,
    Generic,
    RedisValueT,
    TypeVar,
)

from ._key_spec import KeySpec
from .constants import NodeFlag

if TYPE_CHECKING:
    from coredis.cluster._layout import ClusterLayout
    from coredis.cluster._node import ClusterNodeLocation

R = TypeVar("R")


@dataclasses.dataclass(unsafe_hash=True)
class NodeExecution(Generic[R]):
    #: name of the command
    command: bytes
    #: partial arguments derived from the original command
    arguments: tuple[RedisValueT, ...]
    #: Target node
    node: ClusterNodeLocation
    #: The slots the execution affects
    slots: tuple[int, ...] = dataclasses.field(default_factory=tuple)
    #: original indices where the keys of the command were found.
    #: these can be used to reconstruct an ordered response from
    #: multiple nodes
    key_positions: tuple[int, ...] = dataclasses.field(default_factory=tuple, hash=False)
    #: The result for this node execution
    #: (to be populated externally by the executor
    result: R | ResponseError | TimeoutError = dataclasses.field(init=False, hash=False)


class RoutingStrategy(ABC, Generic[R]):
    """
    Base routing strategy that decides how to take a
    request that is not tied to a single slot and
    distribute it to one or many nodes and then
    merge it back into a single response.
    """

    #: Whether this routing strategy supports multiple
    #: slots
    cross_slot: ClassVar[bool] = True

    def __init__(
        self,
        route: NodeFlag | None,
        merge_callback: ClusterMultiNodeCallback[R],
    ) -> None:
        """
        :param route: a hint to which nodes to route to
        :param merge_callback: callback to merge multiple results from
         different nodes into a consistent single response.
        """
        self.route = route
        self.merge_callback = merge_callback

    def combine(
        self,
        node_executions: list[NodeExecution[R]],
    ) -> R:
        return self.merge_callback(
            key_positions=[e.key_positions for e in node_executions],
            responses=[e.result for e in node_executions],
        )

    @property
    @abstractmethod
    def description(
        self,
    ) -> str:
        """
        Description of strategy. Purely for generating documentation
        """
        ...

    @abstractmethod
    def distribute(
        self,
        cluster_layout: ClusterLayout,
        command: bytes,
        arguments: tuple[RedisValueT, ...],
        readonly: bool,
        execution_parameters: ExecutionParameters,
    ) -> list[NodeExecution[R]]:
        """
        Split up the original command into the appropriate sub executions
        """
        ...


class RandomStrategy(RoutingStrategy[R]):
    cross_slot: ClassVar[bool] = False

    def __init__(self) -> None:
        super().__init__(NodeFlag.RANDOM, merge_callback=ClusterNoMerge())

    def distribute(
        self,
        cluster_layout: ClusterLayout,
        command: bytes,
        arguments: tuple[RedisValueT, ...],
        readonly: bool,
        execution_parameters: ExecutionParameters,
    ) -> list[NodeExecution[R]]:
        return [
            NodeExecution(
                node=cluster_layout.random_node(not readonly), command=command, arguments=arguments
            )
        ]

    @property
    def description(self) -> str:
        return NodeFlag.RANDOM.value


class FanoutStrategy(RoutingStrategy[R]):
    cross_slot: ClassVar[bool] = False
    fanout: NodeFlag

    def __init__(
        self,
        route: NodeFlag,
        merge_callback: ClusterMultiNodeCallback[R],
    ):
        super().__init__(route, merge_callback)
        self.fanout = route

    def nodes(
        self,
        cluster_layout: ClusterLayout,
        primary: bool,
        execution_parameters: ExecutionParameters = {},
    ) -> list[ClusterNodeLocation]:
        match self.fanout:
            case NodeFlag.PRIMARIES:
                return cluster_layout.primaries
            case NodeFlag.REPLICAS:
                return cluster_layout.replicas
            case NodeFlag.ALL:
                return cluster_layout.nodes
        raise RuntimeError(f"Invalid route {self.fanout} for fanout strategy")

    def distribute(
        self,
        cluster_layout: ClusterLayout,
        command: bytes,
        arguments: tuple[RedisValueT, ...],
        readonly: bool,
        execution_parameters: ExecutionParameters,
    ) -> list[NodeExecution[R]]:
        return [
            NodeExecution(node=node, command=command, arguments=arguments)
            for node in self.nodes(cluster_layout, not readonly)
        ]

    @property
    def description(self) -> str:
        return self.fanout.value


class SlotRangeStrategy(RoutingStrategy[R]):
    """
    Only for cluster commands that deal with managing slots.
    Commands are routed to appropriate nodes by distributing slot ids in arguments
    to the appropriate nodes that handle them
    """

    def __init__(
        self,
        merge_callback: ClusterMultiNodeCallback[R] = ClusterNoMerge(),
    ):
        super().__init__(NodeFlag.SLOT_ID, merge_callback)

    def distribute(
        self,
        cluster_layout: ClusterLayout,
        command: bytes,
        arguments: tuple[RedisValueT, ...],
        readonly: bool,
        execution_parameters: ExecutionParameters,
    ) -> list[NodeExecution[R]]:
        if slot_arguments_range := execution_parameters.get("slot_arguments_range", None):
            slot_start, slot_end = slot_arguments_range
            arg_slots = arguments[slot_start : slot_end + 1]
            all_slots = list(int(k) for k in arg_slots)
            affected_nodes = cluster_layout.nodes_for_slots(*all_slots)
            node_slots: dict[ClusterNodeLocation, list[int]] = {node: [] for node in affected_nodes}
            for slot in all_slots:
                for node, slots in affected_nodes.items():
                    if slot in slots:
                        node_slots[node].append(slot)
            return [
                NodeExecution[R](
                    slots=tuple(slots),
                    node=node,
                    command=command,
                    arguments=arguments[:slot_start] + tuple(slots) + arguments[slot_end + 1 :],
                )
                for node, slots in node_slots.items()
            ]
        raise RuntimeError("Unable to route {command!r} with arguments {arguments}")

    @property
    def description(self) -> str:
        return NodeFlag.SLOT_ID.value


class KeyRangeStrategy(RoutingStrategy[R]):
    """
    For MGET where responses must be reassembled in the original key order.
    """

    def distribute(
        self,
        cluster_layout: ClusterLayout,
        command: bytes,
        arguments: tuple[RedisValueT, ...],
        readonly: bool,
        execution_parameters: ExecutionParameters,
    ) -> list[NodeExecution[R]]:
        _, key_range = KeySpec.extract_keys(command, *arguments, readonly_command=readonly)
        pre_arguments = arguments[: key_range[0]]
        post_arguments = arguments[key_range[1] + 1 :]
        slots_to_keys = KeySpec.slots_to_keys(command, *arguments, readonly_command=readonly)
        # Track which original positions each node is responsible for
        node_keys: dict[tuple[int, ClusterNodeLocation], list[RedisValueT]] = {}
        node_key_positions: dict[tuple[int, ClusterNodeLocation], list[int]] = {}
        for slot, sub_keys in slots_to_keys.items():
            node = cluster_layout.node_for_slot(slot, primary=not readonly)
            for idx, key in sub_keys:
                node_keys.setdefault((slot, node), []).append(key)
                node_key_positions.setdefault((slot, node), []).append(idx)
        return [
            NodeExecution(
                slots=(slot_node[0],),
                node=slot_node[1],
                command=command,
                arguments=pre_arguments + tuple(node_keys[slot_node]) + post_arguments,
                key_positions=tuple(node_key_positions[slot_node]),
            )
            for slot_node in node_keys
        ]

    @property
    def description(self) -> str:
        return """the nodes serving the keys in the command"""


class PairStrategy(RoutingStrategy[R]):
    """
    For commands such as MSET, MSETEX, MSETNX where key:value pairs are sent as part
    of the request
    """

    def __init__(
        self,
        route: NodeFlag | None,
        merge_callback: ClusterMultiNodeCallback[R],
        key_step: int = 2,
        add_count: bool = False,
    ):
        super().__init__(route, merge_callback)
        self.key_step = key_step
        self.add_count = add_count

    def distribute(
        self,
        cluster_layout: ClusterLayout,
        command: bytes,
        arguments: tuple[RedisValueT, ...],
        readonly: bool,
        execution_parameters: ExecutionParameters,
    ) -> list[NodeExecution[R]]:
        _, key_range = KeySpec.extract_keys(command, *arguments, readonly_command=readonly)
        slots_to_keys = KeySpec.slots_to_keys(command, *arguments, readonly_command=readonly)
        pre_arguments: tuple[RedisValueT, ...] = ()
        if not self.add_count:
            pre_arguments = arguments[: key_range[0]]
        post_arguments = arguments[key_range[1] * self.key_step + 2 :]
        keys = arguments[key_range[0] : key_range[1] * self.key_step + 1 : self.key_step]
        value_section = arguments[key_range[0] + 1 : key_range[1] * self.key_step + self.key_step]
        values = [
            value_section[chunk[0] : chunk[0] + self.key_step - 1]
            for chunk in list(enumerate(value_section))[:: self.key_step]
        ]
        pairs = dict(zip(keys, values))
        node_pairs: dict[tuple[int, ClusterNodeLocation], list[RedisValueT]] = {}
        node_key_positions: dict[tuple[int, ClusterNodeLocation], list[int]] = {}
        for slot, sub_keys in slots_to_keys.items():
            node = cluster_layout.node_for_slot(slot, primary=not readonly)
            for idx, key in sub_keys:
                node_pairs.setdefault((slot, node), []).extend([key, *pairs[key]])
                node_key_positions.setdefault((slot, node), []).append(idx)
            if self.add_count:
                node_pairs[(slot, node)].insert(0, len(sub_keys))
        return [
            NodeExecution(
                slots=(slot_node[0],),
                node=slot_node[1],
                command=command,
                arguments=pre_arguments + tuple(node_pairs) + post_arguments,
                key_positions=tuple(node_key_positions[slot_node]),
            )
            for slot_node, node_pairs in node_pairs.items()
        ]

    @property
    def description(self) -> str:
        return """the nodes serving the keys in the command"""



from __future__ import annotations

import math
import random
import time
from collections import Counter, defaultdict

from anyio import TASK_STATUS_IGNORED, WouldBlock, create_memory_object_stream
from anyio.abc import TaskStatus
from anyio.lowlevel import checkpoint

from coredis._utils import logger, nativestr
from coredis.commands._key_spec import KeySpec
from coredis.commands.constants import CommandName
from coredis.connection import TCPLocation
from coredis.exceptions import (
    ClusterCrossSlotError,
    ConnectionError,
    MovedError,
    RedisClusterError,
    RedisError,
    ResponseError,
)
from coredis.globals import READONLY_COMMANDS
from coredis.retry import ExponentialBackoffRetryPolicy
from coredis.typing import ExecutionParameters, RedisValueT

from ._discovery import DiscoveryService
from ._node import ClusterNodeLocation


class ClusterLayout:
    def __init__(
        self,
        discovery_service: DiscoveryService,
        error_threshold: int = 15,
        maximum_staleness: float = math.inf,
    ) -> None:
        """
        :param discovery_service: The discovery service to use to get the cluster
         layout
        :param error_threshold: Maximum number of errors to observe before forcing
         a refresh of the layout
        :param maximum_staleness: The maximum seconds to tolerate errors while trying
         to refresh the layout. After this threshold, this instance will give up
         trying to keep the layout fresh and the monitor task will raise an error."""
        self._slots: defaultdict[int, list[ClusterNodeLocation]] = defaultdict(list)
        self._nodes: dict[TCPLocation, ClusterNodeLocation] = {}
        self._discovery_service = discovery_service
        self._error_stream = create_memory_object_stream[
            tuple[ClusterNodeLocation | None, Exception]
        ](error_threshold * 2)
        self._error_threshold = error_threshold
        self._last_refresh: float = -math.inf
        self._maximum_staleness = maximum_staleness

    async def initialize(self) -> None:
        await self._refresh()

    def node_for_request(
        self,
        command: bytes,
        arguments: tuple[RedisValueT, ...],
        primary: bool = True,
        execution_parameters: ExecutionParameters = {},
    ) -> ClusterNodeLocation:
        """
        Maps a request to a single node if possible (i.e. all keys
        in the command map to a single slot). This method will not work
        for non keyed commands which must be handled by a routing
        strategy.

        The only exception handled by this method is scripting commands
        (eval/fcall) because they can be called either with or without keys
        and if there are no keys, a random node is as good a choice as any
        """
        if not self._nodes:
            raise RedisClusterError("Local cluster layout cache is empty")

        slots_to_keys = KeySpec.slots_to_keys(
            command, *arguments, readonly_command=command in READONLY_COMMANDS and not primary
        )
        keys, _ = KeySpec.extract_keys(command, *arguments)
        if slots_to_keys:
            if len(slots_to_keys) > 1:
                raise ClusterCrossSlotError(command=command, keys=keys)
            slot = list(slots_to_keys.keys())[0]
            return self.node_for_slot(slot, primary)
        elif command in {
            CommandName.FCALL,
            CommandName.FCALL_RO,
            CommandName.EVAL,
            CommandName.EVAL_RO,
            CommandName.EVALSHA,
            CommandName.EVALSHA_RO,
        }:
            return self.random_node(primary)
        else:
            raise RedisClusterError(
                f"Could not map {nativestr(command)} request to a node in the cluster"
            )

    def node_for_location(self, location: TCPLocation) -> ClusterNodeLocation | None:
        return self._nodes.get(location)

    def node_for_slot(self, slot: int, primary: bool = True) -> ClusterNodeLocation:
        primary_node: ClusterNodeLocation | None = None
        replica_nodes: list[ClusterNodeLocation] = []
        for node in self._slots[slot]:
            if node.is_primary:
                primary_node = node
            else:
                replica_nodes.append(node)
        if primary and primary_node:
            return primary_node
        elif replica_nodes:
            return list(sorted(replica_nodes, key=lambda v: v.priority))[-1]
        # Last resort, if there is no replica, return the primary
        elif primary_node:
            return primary_node
        raise RedisClusterError(
            f"Unable to map slot {slot} to a {'primary' if primary else 'replica'} node"
        )

    def nodes_for_slots(
        self, *slots: int, primary: bool = True
    ) -> dict[ClusterNodeLocation, list[int]]:
        mapping: dict[ClusterNodeLocation, list[int]] = defaultdict(list)
        for slot in slots:
            if node := self.node_for_slot(slot, primary):
                mapping[node].append(slot)
        return mapping

    @property
    def nodes(self) -> list[ClusterNodeLocation]:
        return [node for node in self._nodes.values()]

    @property
    def primaries(self) -> list[ClusterNodeLocation]:
        return [node for node in self._nodes.values() if node.is_primary]

    @property
    def replicas(self) -> list[ClusterNodeLocation]:
        return [node for node in self._nodes.values() if node.server_type == "replica"]

    def random_node(self, primary: bool = True) -> ClusterNodeLocation:
        if primary:
            return random.choice(list(self.primaries))
        else:
            return random.choice(list(self.nodes))

    def _handle_moved_error(
        self,
        error: MovedError,
        errored_node: ClusterNodeLocation,
    ) -> None:
        """
        Updates the node mapping based on a Moved error
        """
        location = TCPLocation(error.host, error.port)
        redirect_node = self._nodes.setdefault(
            location, ClusterNodeLocation(location.host, location.port, server_type="primary")
        )
        slot_nodes = self._slots[error.slot_id]
        for idx, current in enumerate(slot_nodes):
            if (is_errored := (current == errored_node)) or current.is_primary:
                if redirect_node in slot_nodes and is_errored:
                    slot_nodes.pop(idx)
                    break
                else:
                    slot_nodes[idx] = redirect_node

    def report_errors(self, node: ClusterNodeLocation | None, *errors: Exception) -> None:
        for error in errors:
            if node and not isinstance(error, ResponseError):
                node.priority -= 1
            if node and isinstance(error, MovedError):
                self._handle_moved_error(error, node)
            try:
                self._error_stream[0].send_nowait((node, error))
            except WouldBlock:
                pass

    async def monitor(self, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        task_status.started()
        refresh_retry_policy = ExponentialBackoffRetryPolicy(
            (RedisError,),
            retries=None,
            base_delay=1,
            max_delay=max(0, min(30, self._maximum_staleness) - 1),
            deadline=self._maximum_staleness,
        )
        errors: dict[ClusterNodeLocation | None, Counter[type[Exception]]] = defaultdict(Counter)
        while True:
            node, error = await self._error_stream[1].receive()
            errors[node][type(error)] += 1
            total_errors = sum(
                count
                for counter in errors.values()
                for err_type, count in counter.items()
                if issubclass(err_type, (RedisClusterError, ConnectionError))
            )
            if total_errors >= self._error_threshold:
                logger.debug("Refreshing cluster layout due to error threshold exceeded")
                await refresh_retry_policy.call_with_retries(self._refresh)
                errors.clear()

            await checkpoint()

    async def _refresh(self) -> None:
        nodes, slots = await self._discovery_service.get_cluster_layout()
        nodes_by_location = {TCPLocation(node.host, node.port): node for node in nodes}
        self._nodes, self._slots = nodes_by_location, slots
        self._last_refresh = time.monotonic()



from __future__ import annotations

from typing import cast

from coredis.response._callbacks import ResponseCallback
from coredis.typing import (
    AnyStr,
    Iterable,
    ResponsePrimitive,
    ResponseType,
)


class SScanCallback(ResponseCallback[list[ResponseType], tuple[int, set[AnyStr]]]):
    def transform(
        self,
        response: list[ResponseType],
    ) -> tuple[int, set[AnyStr]]:
        cursor, r = response
        assert isinstance(cursor, (bytes, str)) and isinstance(r, Iterable)
        return int(cursor), set(cast(Iterable[AnyStr], r))


class ItemOrSetCallback(
    ResponseCallback[
        AnyStr | list[ResponsePrimitive] | set[ResponsePrimitive],
        AnyStr | set[AnyStr],
    ]
):
    def transform(
        self,
        response: AnyStr | list[ResponsePrimitive] | set[ResponsePrimitive],
    ) -> AnyStr | set[AnyStr]:
        if self.options.get("count"):
            if isinstance(response, set):
                return cast(set[AnyStr], response)
            if isinstance(response, list):
                return cast(set[AnyStr], set(response) if response else set())
            raise ValueError(f"Unable to map {response!r} to set")
        else:
            return cast(AnyStr, response)
